﻿using ESRI.ArcGIS.Client.Geometry;
using ESRI.ArcGIS.Client.Projection;
using Microsoft.SqlServer.Types;
using System;
using System.Collections.Generic;
using System.Data.Spatial;
using System.Data.SqlTypes;
using System.Globalization;

namespace Sigcat.Client.Model
{
    public class GeographyWrapper
    {
        #region Constants

        /// <summary>
        /// Identificador del sistema de referència de coordenades geodèsiques
        /// ETRS89 UTM fus 31 (25831).
        /// </summary>
        public const int SRID_ETRS89_UTM_FUS31 = 25831;

        /// <summary>
        /// Identificador del sistema de referència de coordenades geodèsiques
        /// WGS84 (4326) que ESRI utilitza per defecte pel sistema geogràfic.
        /// </summary>
        public const int SRID_WGS84_GeographicDefaultByESRI = 4326;

        #endregion

        #region Members

        /// <summary>
        /// Well-Known Text
        /// </summary>
        private string wkt;

        #endregion

        #region Properties

        #region Property WKT

        /// <summary>
        /// Representació de l'element geodèsic de tipus cadena mitjançant l'estàndard <remarks>Well-Known Text</remarks>.
        /// </summary>
        public string WKT
        {
            get { return wkt; }
            set { wkt = value; }
        }

        #endregion

        #region Property dbGeography

        /// <summary>
        /// Representació de l'element geodèsic de tipus <see cref="DbGeography"/>.
        /// </summary>
        public DbGeography dbGeography
        {
            get { return DbGeography.FromText(WKT); }
            set { WKT = value.AsText(); }
        }

        #endregion

        #region Property geometry

        /// <summary>
        /// Representació de l'element geodèsic de tipus <see cref="Geometry">ESRI Geometry</see>.
        /// </summary>
        public Geometry geometry
        {
            get { return GeometryFromWKT(WKT); }
            set { WKT = GeometryToWKT(value); }
        }

        #endregion

        #endregion

        #region Constructors

        /// <summary>
        /// Constructor a partir d'una cadena coneguda <remarks>Well-Known Text</remarks>.
        /// </summary>
        /// <param name="WKT">Cadena coneguda (Well-Known Text) amb la representació geodèsica.</param>
        public GeographyWrapper(string WKT)
        {
            this.WKT = WKT;
        }

        /// <summary>
        /// Constructor a partir d'un objecte <see cref="DbGeography">DbGeography</see>.
        /// </summary>
        /// <param name="g">Objecte <see cref="DbGeography">DbGeography</see></param>.
        public GeographyWrapper(DbGeography g)
        {
            this.dbGeography = g;
        }

        /// <summary>
        /// Constructor a partir d'un objecte <see cref="Geometry">ESRI Geometry</see>.
        /// </summary>
        /// <param name="g">Objecte <see cref="Geometry">ESRI Geometry</see>.
        public GeographyWrapper(Geometry g)
        {
            this.geometry = g;
        }

        #endregion

        #region Operators

        #region Operator From/To string

        /// <summary>
        /// Assigna un objecte <see cref="GeographyWrapper"/> a partir d'una cadena coneguda <remarks>Well-Known Text</remarks>.
        /// </summary>
        /// <param name="WKT">Cadena coneguda (Well-Known Text) amb la representació geodèsica.</param>
        /// <returns>Objecte <see cref="GeographyWrapper"/>.</returns>
        public static implicit operator GeographyWrapper(string WKT)
        {
            return new GeographyWrapper(WKT);
        }

        /// <summary>
        /// Assigna una cadena coneguda <remarks>Well-Known Text</remarks> a partir d'un objecte <see cref="GeographyWrapper"/>.
        /// </summary>
        /// <param name="gw">Objecte <see cref="GeographyWrapper"/>.</param>
        /// <returns>Cadena coneguda (Well-Known Text) amb la representació geodèsica.</returns>
        public static implicit operator string(GeographyWrapper gw)
        {
            return gw.wkt;
        }

        #endregion

        #region Operator From/To DbGeography

        /// <summary>
        /// Assigna un objecte <see cref="GeographyWrapper"/> a partir d'un objecte <see cref="DbGeography"/>.
        /// </summary>
        /// <param name="g">Objecte <see cref="DbGeography"/>.</param>
        /// <returns>Objecte <see cref="GeographyWrapper"/>.</returns>
        public static implicit operator GeographyWrapper(DbGeography g)
        {
            return new GeographyWrapper(g);
        }

        /// <summary>
        /// Assigna un objecte <see cref="DbGeography"/> a partir d'un objecte <see cref="GeographyWrapper"/>.
        /// </summary>
        /// <param name="gw">Objecte <see cref="GeographyWrapper"/>.</param>
        /// <returns>Objecte <see cref="DbGeography"/>.</returns>
        public static implicit operator DbGeography(GeographyWrapper gw)
        {
            return gw.dbGeography;
        }

        #endregion

        #region Operator From/To ESRI Geometry

        /// <summary>
        /// Assigna un objecte <see cref="GeographyWrapper"/> a partir d'un objecte <see cref="Geometry">ESRI Geometry</see>.
        /// </summary>
        /// <param name="g">Objecte <see cref="Geometry">ESRI Geometry</see>.</param>
        /// <returns>Objecte <see cref="GeographyWrapper"/>.</returns>
        public static implicit operator GeographyWrapper(Geometry g)
        {
            return new GeographyWrapper(g);
        }

        /// <summary>
        /// Assigna un objecte <see cref="Geometry">ESRI Geometry</see> a partir d'un objecte <see cref="GeographyWrapper"/>.
        /// </summary>
        /// <param name="gw">Objecte <see cref="GeographyWrapper"/>.</param>
        /// <returns>Objecte <see cref="Geometry">ESRI Geometry</see>.</returns>
        public static implicit operator Geometry(GeographyWrapper gw)
        {
            return gw.geometry;
        }

        #endregion

        #endregion

        #region Static Methods

        #region Static Method GeometryFromWKT

        /// <summary>
        /// Converteix una cadena coneguda <remarks>Well-Known Text</remarks>
        /// en un objecte <see cref="Geometry">ESRI Geometry</see>.
        /// </summary>
        /// <param name="WKT">Cadena coneguda (Well-Known Text) amb la representació geodèsica.</param>
        /// <returns>Objecte <see cref="Geometry">ESRI Geometry</see>.</returns>
        public static Geometry GeometryFromWKT(string WKT)
        {
            try
            {
                DbGeography g = DbGeography.FromText(WKT);
                switch (g.SpatialTypeName.ToUpper())
                {
                    case "POINT":
                        // Es crea un objecte ESRI MapPoint que hereda de ESRI Geometry.
                        MapPoint esriPoint = new MapPoint(g.Longitude.Value, g.Latitude.Value,
                                new SpatialReference(g.CoordinateSystemId));
                        // Es retorna el punt.
                        return esriPoint;
                    case "MULTIPOINT":
                        break;
                    case "MULTILINESTRING":
                        // Es crea un objecte ESRI Polyline que hereda de ESRI Geometry.
                        Polyline esriPolyline = new Polyline();
                        // S'estableix el sistema de espacial de referència.
                        esriPolyline.SpatialReference = new SpatialReference(g.CoordinateSystemId);
                        // Es recorren totes les línies:
                        for (int k = 1; k <= g.ElementCount; k++)
                        {
                            DbGeography line = g.ElementAt(k);
                            // S'obtenen tots els punts de la línia:
                            PointCollection esriLine_points = new PointCollection();
                            for (int l = 1; l <= line.PointCount; l++)
                            {
                                DbGeography esriPoly_point = line.PointAt(l);
                                esriLine_points.Add(new MapPoint(esriPoly_point.Longitude.Value, esriPoly_point.Latitude.Value,
                                    new SpatialReference(esriPoly_point.CoordinateSystemId)));
                            }
                            // S'afegeixen els punts que encadenen les línies.
                            esriPolyline.Paths.Add(esriLine_points);
                        }
                        // Es retornen les línies.
                        return esriPolyline;
                    case "POLYGON":
                    case "MULTIPOLYGON":
                        // Es crea un objecte ESRI Polygon que hereda de ESRI Geometry.
                        Polygon esriPolygon = new Polygon();
                        // S'estableix el sistema de espacial de referència.
                        esriPolygon.SpatialReference = new SpatialReference(g.CoordinateSystemId);
                        // Es recorren tots els polígons:
                        for (int k = 1; k <= g.ElementCount; k++)
                        {
                            DbGeography polygon = g.ElementAt(k);
                            // S'obtenen tots els punts del polígon:
                            PointCollection esriPoly_points = new PointCollection();
                            for (int l = 1; l <= polygon.PointCount; l++)
                            {
                                DbGeography esriPoly_point = polygon.PointAt(l);
                                esriPoly_points.Add(new MapPoint(esriPoly_point.Longitude.Value, esriPoly_point.Latitude.Value,
                                    new SpatialReference(esriPoly_point.CoordinateSystemId)));
                            }
                            // S'afegeixen els punts com anell del polígon.
                            esriPolygon.Rings.Add(esriPoly_points);
                        }
                        // Es retorna el polígon.
                        return esriPolygon;
                    default:
                        break;
                }
            }
            catch (Exception e)
            {
                // En cas d'error es captura per que no es propagui fora de la funció.
            }
            return null;
        }

        #endregion

        #region Static Method GeometryToWKT

        /// <summary>
        /// Converteix un objecte <see cref="Geometry">ESRI Geometry</see>
        /// en una cadena coneguda estàndard <remarks>Well-Known Text</remarks>.
        /// </summary>
        /// <param name="g">Objecte <see cref="Geometry">ESRI Geometry</see>.</param>
        /// <returns>Cadena coneguda (Well-Known Text) amb la representació geodèsica.</returns>
        public static string GeometryToWKT(Geometry g)
        {
            // Si l'objecte geomètric és nul es retorna nul.
            if (g == null) return null;

            // Si l'objecte geomètric no es troba en coordenades geogràfiques es transforma.
            if (g.SpatialReference.WKID != SRID_WGS84_GeographicDefaultByESRI)
                g = new WebMercator().ToGeographic(g);

            string wkt; // Cadena coneguda (Well-Known Text) de l'objecte geomètric.

            // S'utilitza el sistema americà per donar format a les cadenes de text per a que 
            // els nombres decimals es representin amb punt decimal i no amb coma decimal.
            IFormatProvider fp = CultureInfo.GetCultureInfo("en-US");

            if (g is MapPoint)
            // En cas que es tracti d'un punt:
            {
                MapPoint point = g as MapPoint;
                wkt = string.Format(fp, "POINT({0} {1})", point.X, point.Y);
            }
            else if (g is MultiPoint)
                // En cas que es tracti d'un conjunt de punts:
            {
                MultiPoint mpoint = g as MultiPoint;
                List<string> wktPoints = new List<string>();
                foreach (MapPoint point in mpoint.Points)
                    wktPoints.Add(string.Format(fp, "{0} {1}", point.X, point.Y));
                wkt = string.Format("MULTIPOINT({0})", string.Join(", ", wktPoints));
            }
            else if (g is Polyline)
                // En cas que es tracti d'un conjunt de linies:
            {
                Polyline line = g as Polyline;
                List<string> wktPaths = new List<string>();
                foreach (PointCollection points in line.Paths)
                {
                    List<string> wktPoints = new List<string>();
                    foreach (MapPoint point in points)
                        wktPoints.Add(string.Format(fp, "{0} {1}", point.X, point.Y));
                    wktPaths.Add(string.Format("({0})", string.Join(", ", wktPoints)));
                }
                wkt = string.Format("MULTILINESTRING({0})", string.Join(", ", wktPaths));
            }
            else if (g is Polygon)
                // En cas que es tracti d'un polígon:
            {
                Polygon polygon = g as Polygon;
                List<string> wktPaths = new List<string>();
                foreach (PointCollection points in polygon.Rings)
                {
                    List<string> wktPoints = new List<string>();
                    foreach (MapPoint point in points)
                        wktPoints.Add(string.Format(fp, "{0} {1}", point.X, point.Y));
                    wktPaths.Add(string.Format("({0})", string.Join(", ", wktPoints)));
                }
                wkt = string.Format("POLYGON({0})", string.Join(", ", wktPaths));
                // Es converteix el polígon en un polígon vàlid si les arestes s'entrecreuen,
                // s'utilitza la funcionalitat MakeValid de SQL que ho implementa.
                // SQL pot convertir el POLYGON en MULTIPOLYGON
                var sqlGeography = SqlGeography.STPolyFromText(new SqlChars(wkt), SRID_WGS84_GeographicDefaultByESRI).MakeValid();
                // En cas que l'ordre dels punts del polígon sigui incorrecte es podria prendre com a vàlid
                // un polígon invertit, per aquest motiu cal agafar el polígon de menor àrea.
                var sqlGeography_inverted = sqlGeography.ReorientObject();
                if (sqlGeography.STArea() > sqlGeography_inverted.STArea())
                    sqlGeography = sqlGeography_inverted;
                // s'obté la representació WKT del polígon o multipolígon.
                wkt = sqlGeography.STAsText().ToSqlString().Value;
            }
            else if (g is Envelope)
                // En cas que es tracti d'un envoltori (rectangle):
            {
                Envelope envelope = g as Envelope;
                wkt = string.Format(fp, "POLYGON(({0} {1}, {2} {1}, {2} {3}, {0} {3}, {0} {1}))",
                    envelope.XMin, envelope.YMin, envelope.XMax, envelope.YMax);
            }
            else
                // En cas contrari:
            {
                // es retorna nul.
                return null;
            }
            // Es retorna la cadena coneguda (Well-Known Text) amb la representació geodèsica.
            return wkt;
        }

        #endregion

        #endregion

    }

}
