﻿using Newtonsoft.Json;
using System;
using System.IO;
using System.Threading.Tasks;
using System.Xml;

namespace Sigcat.Client.Model.Settings
{

    public class DataServiceSettings : DataServiceBase, IDataServiceSettings
    {

        #region Members

        /// <summary>
        /// Nom i extensió del fitxer de configuració de l'aplicació.
        /// </summary>
        private const string SettingsFileName = "settings.xml";

        /// <summary>
        /// Nom de l'element Xml arrel del fitxer de configuració.
        /// </summary>
        private const string RootXmlElement = "Settings";

        /// <summary>
        /// Caché a nivell d'aplicació del model de dades <see cref="SettingsModel"/>.
        /// </summary>
        private static SettingsModel settingsModel;

        #endregion

        #region Service Task<SettingsModel> GetSettings()

        /// <summary>
        /// Obté el model <see cref="SettingsModel"/> de l'aplicació.
        /// </summary>
        /// <returns>
        /// Tasca amb el model de dades
        /// <see cref="SettingsModel"/> de l'aplicació.
        /// </returns>
        /// <remarks>
        /// Si no s'ha obtingut la configuració abans
        /// es recupera del fitxer de configuració.
        /// </remarks>
        public async Task<SettingsModel> GetSettings()
        {
            if (settingsModel == null)
                settingsModel = await RestoreSettings();
            return settingsModel;
        }

        #endregion

        #region Service Task<SettingsModel> RestoreSettings()

        /// <summary>
        /// Obté el model <see cref="SettingsModel"/> de l'arxiu XML
        /// de configuració de l'aplicació.
        /// </summary>
        /// <returns>
        /// Tasca amb el model de dades
        /// <see cref="SettingsModel"/> de l'aplicació.
        /// </returns>
        /// <remarks>
        /// Si no es pot recuperar el fitxer es retornen dades per defecte.
        /// </remarks>
        public Task<SettingsModel> RestoreSettings()
        {
            // Obté la configuració del fitxer de l'aplicació.
            try
            {
                // S'obté el document XML del fitxer de configuració.
                string xmlText = File.ReadAllText(SettingsFileName);
                var doc = new XmlDocument();
                doc.LoadXml(xmlText);
                // Es serialitza el document XML a Json.
                string json = JsonConvert.SerializeXmlNode(doc.DocumentElement,
                    Newtonsoft.Json.Formatting.Indented, true);
                // Es deserialitza Json al model de dades SettingsModel.
                settingsModel = JsonConvert.DeserializeObject<SettingsModel>(json);
                // Si no s'han obtingut les dades correctament es retornen dades per defecte.
                if (settingsModel == null)
                    throw new Exception();
            }
            catch (Exception)
            {
                // En cas d'error es retorna un model de dades per defecte.
                settingsModel = new SettingsModel()
                {
                    AuthenticationService = new ServiceModel() { Path = "Authentication.svc" },
                    DomainService = new ServiceModel() { Path = "Domain.svc" }
                };
            }
            return Task.FromResult(settingsModel);
        }

        #endregion

        #region Service Task SaveSettings(SettingsModel settings)

        /// <summary>
        /// Emmagatzema el model de dades <see cref="SettingsModel"/>
        /// al fitxer de configuració Xml de l'aplicació.
        /// </summary>
        /// <param name="settings">
        /// Model de dades <see cref="SettingsModel"/>
        /// </param>
        /// <returns>Tasca sense resultat.</returns>
        public async Task SaveSettings(SettingsModel settings)
        {
            string json = await JsonConvert.SerializeObjectAsync(settings);
            XmlDocument doc = JsonConvert.DeserializeXmlNode(json, RootXmlElement);
            string result = await JsonConvert.SerializeObjectAsync(doc);
            doc.Save(SettingsFileName);
        }

        #endregion

        #region Service Task<bool> CheckService(ServiceTypes serviceType)

        /// <summary>
        /// Comprova la connexió amb el servei <see cref="ServiceModel"/>.
        /// </summary>
        /// <param name="serviceType">
        /// Servei de l'aplicació al que es refereix el model de dades.
        /// </param>
        /// <returns>
        /// Vertader si es pot realitzar la connexió amb el servei.
        /// Fals si no es pot connectar amb el servei.
        /// </returns>
        public async Task<bool> CheckService(ServiceTypes serviceType)
        {
            // Captura les possibles excepcions.
            try
            {
                // Obté el model del servei.
                ServiceModel service = await GetServiceModel(serviceType);
                // Obté l'adreça url del servei.
                string httpGetUrl = await GetServiceUri(serviceType);
                // Obté la tecnologia del servei utilitzada.
                EndPoints endPoint = service.EndPoint;
                // Realitza la connexió segons la tecnologia.
                switch (endPoint)
                {
                    case EndPoints.OData:
                        /* -------------------------> FALTA IMPLEMENTAR!
                        // Connecta amb el servei concret de l'aplicació.
                        switch (serviceType)
                        {
                            case ServiceTypes.AuthSvc:
                                break;
                            case ServiceTypes.DomSvc:
                            default:
                                break;
                        }*/
                        break;
                    case EndPoints.Soap:
                        /* -------------------------> FALTA IMPLEMENTAR!
                        // Connecta amb el servei concret de l'aplicació.
                        switch (serviceType)
                        {
                            case ServiceTypes.AuthSvc:
                                break;
                            case ServiceTypes.DomSvc:
                            default:
                                break;
                        }*/
                        break;
                    case EndPoints.Json:
                    default:
                        httpGetUrl = string.Format("{0}/Test", httpGetUrl);
                        TestServiceResult result = await GetJsonObject<TestServiceResult>(httpGetUrl);
                        // S'obté la resposta del servidor a aquesta adreça.
                        if (result == null)
                            return false;
                        else
                            return result.Tested;
                }
                return false;
            }
            catch (Exception)
            {
                // Per qualsevol excepció es retorna fals.
                return false;
            }
        }

        #endregion

        #region Service Task<string> GetServiceUri(ServiceTypes serviceType)

        /// <summary>
        /// Obté la cadena de connexió del servei <see cref="ServiceModel"/>.
        /// </summary>
        /// <param name="serviceType">
        /// Servei de l'aplicació al que es refereix el model de dades.
        /// </param>
        /// <returns>
        /// Cadena de connexió Uri del servei.
        /// </returns>
        public async Task<string> GetServiceUri(ServiceTypes serviceType)
        {
            ServiceModel service = await GetServiceModel(serviceType);
            string httpGetUrl = GetServiceUri(service);
            return httpGetUrl;
        }

        #endregion

        #region Method string GetServiceUri(ServiceModel service)

        /// <summary>
        /// Obté la cadena de connexió del servei <see cref="ServiceModel"/>.
        /// </summary>
        /// <param name="ServiceModel">
        /// Model de dades d'un servei de l'aplicació.
        /// </param>
        /// <returns>
        /// Cadena de connexió Uri del servei.
        /// </returns>
        /// <remarks>
        /// Si es produeix una excepció es retorna l'excepció.
        /// </remarks>
        public static string GetServiceUri(ServiceModel service)
        {
            try
            {
                // Es crea un constructor d'adreces UriBuider amb el model del servei.
                UriBuilder ub = new UriBuilder()
                {
                    Scheme = Enum.GetName(typeof(SchemeTypes), service.Scheme),
                    Host = service.Host,
                    Port = service.Port,
                    Path = service.Path
                };
                // S'obté l'adreça url del servei.
                string httpGetUrl;
                switch (service.EndPoint)
                {
                    case EndPoints.OData:
                        httpGetUrl = string.Format("{0}/odata", ub.Uri.ToString());
                        break;
                    case EndPoints.Json:
                        httpGetUrl = string.Format("{0}/json", ub.Uri.ToString());
                        break;
                    case EndPoints.Soap:
                    default:
                        httpGetUrl = ub.Uri.ToString();
                        break;
                }
                return httpGetUrl;
            }
            catch (Exception ex)
            {
                // En cas d'excepció es retorna l'excepció.
                return ex.Message;
            }
        }

        #endregion

        #region Method Task<ServiceModel> GetServiceModel(ServiceTypes serviceType)

        /// <summary>
        /// Obté el model d'un servei de l'aplicació.
        /// </summary>
        /// <param name="serviceType">Tipus de servei.</param>
        /// <returns>Servei de l'aplicació.</returns>
        public static async Task<ServiceModel> GetServiceModel(ServiceTypes serviceType)
        {
            DataServiceSettings ds = new DataServiceSettings();
            SettingsModel settings = await ds.GetSettings();
            ServiceModel service = GetServiceModel(serviceType, settings);
            return service;
        }

        #endregion

        #region Method ServiceModel GetServiceModel(ServiceTypes serviceType, SettingsModel settings)

        /// <summary>
        /// Obté el model d'un servei del model de la configuració de l'aplicació.
        /// </summary>
        /// <param name="serviceType">Tipus de servei.</param>
        /// <param name="settings">Model de la configuració.</param>
        /// <returns>Servei de l'aplicació.</returns>
        public static ServiceModel GetServiceModel(ServiceTypes serviceType, SettingsModel settings)
        {
            ServiceModel service;
            switch (serviceType)
            {
                case ServiceTypes.AuthSvc:
                    service = settings.AuthenticationService;
                    break;
                case ServiceTypes.DomSvc:
                default:
                    service = settings.DomainService;
                    break;
            }
            return service;
        }

        #endregion

    }

}
