﻿using fgxavi.Mvvm;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Messaging;
using Sigcat.Client.Model.Settings;
using Sigcat.Client.ViewModel.Resources;
using System;
using System.ComponentModel;
using System.Linq.Expressions;

namespace Sigcat.Client.ViewModel.Settings
{
    public class SettingsViewModel : ViewModelBase
    {

        #region Members

        #region Member IDataServiceSettings

        /// <summary>
        /// Membre privat amb accés al servei del model de dades.
        /// </summary>
        private readonly IDataServiceSettings _svc;

        #endregion

        #endregion

        #region Properties

        #region Property Settings

        /// <summary>
        /// Membre privat de la propietat <see cref="Settings" />.
        /// </summary>
        private SettingsModel _Settings = null;

        /// <summary>
        /// Estableix i obté la propietat <see cref="Settings" />.
        /// Fa ús del mètode <see cref="SetField"/> per establir la propietat.
        /// </summary>
        public SettingsModel Settings
        {
            get { return _Settings; }
            set { SetField(ref _Settings, value); }
        }

        #endregion

        #endregion

        #region Commands

        #region RelayCommand<ServiceModel> VerifyCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="VerifyCommand" />.
        /// </summary>
        private RelayCommand<ServiceModel> _VerifyCommand;

        /// <summary>
        /// Obté la commanda <see cref="VerifyCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<ServiceModel> VerifyCommand
        {
            get
            {
                return _VerifyCommand
                    ?? (_VerifyCommand = new RelayCommand<ServiceModel>(
                        ExecuteVerifyCommand,
                        CanExecuteVerifyCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="VerifyCommand" />.
        /// </summary>
        /// <param name="service">Paràmetre de la commanda.</param>
        private async void ExecuteVerifyCommand(ServiceModel service)
        {
            // No s'executa en mode diseny.
            if (IsInDesignMode)
                return;
            service.Verified = VerifiedStatus.Establishing;
            try
            {
                bool result = await _svc.CheckService(service.ServiceType);
                if (result)
                    service.Verified = VerifiedStatus.Verified;
                else
                    service.Verified = VerifiedStatus.Failed;
            }
            catch (Exception)
            {
                service.Verified = VerifiedStatus.Failed;
            }
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="VerifyCommand" />.
        /// </summary>
        /// <param name="service">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        private bool CanExecuteVerifyCommand(ServiceModel service)
        {
            return service != null ? !service.HasErrors : false;
        }

        #endregion

        #region RelayCommand SaveXmlCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="SaveXmlCommand" />.
        /// </summary>
        private RelayCommand _SaveXmlCommand;

        /// <summary>
        /// Obté la commanda <see cref="SaveXmlCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand SaveXmlCommand
        {
            get
            {
                return _SaveXmlCommand
                    ?? (_SaveXmlCommand = new RelayCommand(
                        ExecuteSaveXmlCommand,
                        CanExecuteSaveXmlCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="SaveXmlCommand" />.
        /// </summary>
        private async void ExecuteSaveXmlCommand()
        {
            await _svc.SaveSettings(Settings);
            Settings.DomainService.AcceptChanges();
            Settings.AuthenticationService.AcceptChanges();
            Settings.AcceptChanges();
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="SaveXmlCommand" />.
        /// </summary>
        /// <returns>Autorització d'execució de la commanda.</returns>
        private bool CanExecuteSaveXmlCommand()
        {
            return !Settings.HasErrors
                && !Settings.AuthenticationService.HasErrors
                && !Settings.DomainService.HasErrors
                && (Settings.AuthenticationService.IsChanged || Settings.DomainService.IsChanged);
        }

        #endregion

        #region RelayCommand RestoreXmlCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RestoreXmlCommand" />.
        /// </summary>
        private RelayCommand _RestoreXmlCommand;

        /// <summary>
        /// Obté la commanda <see cref="RestoreXmlCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand RestoreXmlCommand
        {
            get
            {
                return _RestoreXmlCommand
                    ?? (_RestoreXmlCommand = new RelayCommand(
                        ExecuteRestoreXmlCommand,
                        CanExecuteRestoreXmlCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RestoreXmlCommand" />.
        /// </summary>
        private async void ExecuteRestoreXmlCommand()
        {
            Settings = await _svc.RestoreSettings();
            Settings.AuthenticationService.PropertyChanged += Service_PropertyChanged;
            Settings.DomainService.PropertyChanged += Service_PropertyChanged;
            Service_ChangeUrl(Settings.AuthenticationService);
            Service_ChangeUrl(Settings.DomainService);
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RestoreXmlCommand" />.
        /// </summary>
        /// <returns>Autorització d'execució de la commanda.</returns>
        private bool CanExecuteRestoreXmlCommand()
        {
            return CanExecuteSaveXmlCommand();
        }

        #endregion

        #endregion

        #region Contructor

        /// <summary>
        /// Contructor per defecte.
        /// </summary>
        public SettingsViewModel(IDataServiceSettings svc)
        {
            // Afegeix una referència al servei del model.
            _svc = svc;

            // Obté el model de dades de servei.
            ExecuteRestoreXmlCommand();
        }

        #endregion

        #region Method Service_PropertyChanged

        /// <summary>
        /// Controlador de l'event PropertyChanged.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void Service_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            ServiceModel service = sender as ServiceModel;
            switch (e.PropertyName)
            {
                case "Verified":
                case "Url":
                    break;
                default:
                    Service_ChangeUrl(service);
                    break;
            }
        }

        #endregion

        #region Method Service_PropertyChanged

        /// <summary>
        /// Canvia la propietat Url de <see cref="ServiceModel"/>.
        /// </summary>
        /// <param name="service">Servei de l'aplicació.</param>
        protected void Service_ChangeUrl(ServiceModel service)
        {
            try
            {
                if (service.HasErrors)
                    throw new Exception();
                UriBuilder ub = new UriBuilder()
                {
                    Scheme = Enum.GetName(typeof(SchemeTypes), service.Scheme),
                    Host = service.Host,
                    Port = service.Port,
                    Path = service.Path
                };
                service.Url = ub;
            }
            catch (Exception)
            {
                service.Url = null;
            }
        }

        #endregion

    }

}
