using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Ioc;
using Microsoft.Practices.ServiceLocation;
using Sigcat.Client.Model;
using Sigcat.Client.Model.Directorate;
using Sigcat.Client.Model.Settings;
using Sigcat.Client.ViewModel.Directorate;
using Sigcat.Client.ViewModel.Settings;

namespace Sigcat.Client.ViewModel
{
    /// <summary>
    /// This class contains static references to all the view models in the
    /// application and provides an entry point for the bindings.
    /// </summary>
    public class ViewModelLocator
    {

        #region Constructor

        /// <summary>
        /// Initializes a new instance of the ViewModelLocator class.
        /// </summary>
        public ViewModelLocator()
        {
            ServiceLocator.SetLocatorProvider(() => SimpleIoc.Default);

            if (ViewModelBase.IsInDesignModeStatic)
            {
                // Create design time view services and models
                SimpleIoc.Default.Register<IDataServiceSettings, DesignDataServiceSettings>();
                SimpleIoc.Default.Register<IRepository<DirectorateModel>, DesignRepositoryDirectorate>();
            }
            else
            {
                // Create run time view services and models
                SimpleIoc.Default.Register<IDataServiceSettings, DataServiceSettings>();
                SimpleIoc.Default.Register<IRepository<DirectorateModel>, Repository<DirectorateModel>>();
            }

            SimpleIoc.Default.Register<SettingsViewModel>();
            SimpleIoc.Default.Register<DirectorateViewModel>();
            SimpleIoc.Default.Register<EsriMapViewModel>();
            SimpleIoc.Default.Register<EsriSymbolSelectorViewModel>();
        }

        #endregion

        #region Property SettingsViewModel

        /// <summary>
        /// Gets the Settings property.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance",
            "CA1822:MarkMembersAsStatic",
            Justification = "This non-static member is needed for data binding purposes.")]
        public SettingsViewModel Settings
        {
            get
            {
                return ServiceLocator.Current.GetInstance<SettingsViewModel>();
            }
        }

        #endregion

        #region DirectorateViewModel

        /// <summary>
        /// Gets the Directorates property.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance",
            "CA1822:MarkMembersAsStatic",
            Justification = "This non-static member is needed for data binding purposes.")]
        public DirectorateViewModel Directorates
        {
            get
            {
                return ServiceLocator.Current.GetInstance<DirectorateViewModel>();
            }
        }

        #endregion

        #region Property EsriMapViewModel

        /// <summary>
        /// Gets the EsriMap property.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance",
            "CA1822:MarkMembersAsStatic",
            Justification = "This non-static member is needed for data binding purposes.")]
        public EsriMapViewModel EsriMap
        {
            get
            {
                return ServiceLocator.Current.GetInstance<EsriMapViewModel>();
            }
        }

        #endregion

        #region Property EsriSymbolSelectorViewModel

        /// <summary>
        /// Gets the EsriSymbolSelector property.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance",
            "CA1822:MarkMembersAsStatic",
            Justification = "This non-static member is needed for data binding purposes.")]
        public EsriSymbolSelectorViewModel EsriSymbolSelector
        {
            get
            {
                return ServiceLocator.Current.GetInstance<EsriSymbolSelectorViewModel>();
            }
        }

        #endregion

        // more mvvmlocatorproperty

    }

}
