﻿using Sigcat.Server.Helpers;
using Sigcat.Server.Model;
using System;
using System.Linq;
using System.ServiceModel.DomainServices.EntityFramework;
using System.ServiceModel.DomainServices.Hosting;
using System.ServiceModel.DomainServices.Server;
using System.ServiceModel.DomainServices.Server.ApplicationServices;
using System.Web;
using System.Web.Security;

namespace Sigcat.Server.WcfSvcLib
{
    [EnableClientAccess]
    public class AuthSvc : LinqToEntitiesDomainService<SigcatEntities>, IAuthentication<SessionUser>, ITestService
    {

        #region ITestService implementation

        /// <summary>
        /// Verifica que el servei web és correcte.
        /// Serveix per realitzar proves de connexió.
        /// </summary>
        /// <returns>
        /// Un objecte <see cref="TestServiceResult"/>
        /// que confirma que s'ha realitzat la connexió.
        /// </returns>
        public TestServiceResult Test()
        {
            var result = new TestServiceResult()
            {
                ServiceName = this.GetType().Name,
                Tested = true
            };
            return result;
        }

        #endregion

        #region Private static member AnonymousUser

        /// <summary>
        /// Sessió d'un usuari anònim.
        /// </summary>
        private static readonly SessionUser AnonymousUser = new SessionUser
        {
            Name = "Anònim",
        };

        #endregion

        #region Methods IAuthentication<LoginUser> implementation

        #region Method IAuthentication<LoginUser>.GetUser

        /// <summary>
        /// Obté la sessió de l'usuari actual.
        /// </summary>
        /// <returns>
        /// Sessió de l'usuari actual.
        /// </returns>
        [Query(IsComposable = false)]
        public SessionUser GetUser()
        {
            // Si l'usuari està autentificat s'obté la seva sessió.
            if ((ServiceContext != null) &&
                (ServiceContext.User != null) &&
                ServiceContext.User.Identity.IsAuthenticated)
            {
                try
                {
                    // Es converteix l'etiqueta de l'estat de sessió al tipus utilitzat.
                    FormsIdentity id = ServiceContext.User.Identity as FormsIdentity;
                    string username = id.Name;
                    int directorateId = int.Parse(id.Ticket.UserData);
                    // Es retorna la sessió del usuari.
                    return GetDirectorateUserByName(directorateId, username);
                }
                catch (Exception)
                {
                    // Si es produeix qualsevol excepció es retorn l'usuari anònim.
                    return AnonymousUser;
                }
            }
            // Si l'usuari no està autentificat es retorna l'usuari anònim.
            return AnonymousUser;
        }

        #endregion

        #region Method IAuthentication<LoginUser>.UpdateUser

        /// <summary>
        /// Not implemented.
        /// </summary>
        /// <param name="user"></param>
        [Update]
        public void UpdateUser(SessionUser user)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region Method IAuthentication<LoginUser>.Login

        /// <summary>
        /// Validate and login
        /// </summary>
        /// <param name="userName"></param>
        /// <param name="password"></param>
        /// <param name="isPersistent"></param>
        /// <param name="directorateIdStr"></param>
        /// <returns></returns>
        public SessionUser Login(string userName, string password, bool isPersistent, string directorateIdStr)
        {
            // Sessió d'usuari a retornar.
            SessionUser foundUser;
            try
            {
                // S'obté el codi de la direcció general.
                int directorateId = int.Parse(directorateIdStr);

                // Si l'usuari és vàlid s'obté l'usuari i es crea el seu estat de sessió.
                if (ValidateDirectorateUser(directorateId, userName, () => password, out foundUser))
                {
                    // Es crea l'estat de sessió de l'usuari.
                    FormsAuthenticationTicket ticket = new FormsAuthenticationTicket(
                        1, // versió de l'estat de sessió
                        userName,
                        // si la sessió és persistent es manté durant mitja hora o fins que es realitzi logout
                        DateTime.Now, DateTime.Now.AddMinutes(30),
                        isPersistent,
                        directorateIdStr, // s'emmagatzema a l'estat de sessió el codi de la direcció general
                        FormsAuthentication.FormsCookiePath);

                    // S'encripta l'estat de la sessió.
                    string encryptedTicket = FormsAuthentication.Encrypt(ticket);

                    // Es crea l'etiqueta de la sessió.
                    HttpCookie authCookie = new HttpCookie(FormsAuthentication.FormsCookieName, encryptedTicket);

                    // Si la sessió és persistent s'afegeix data de finalització a l'estat de sessió.
                    if (ticket.IsPersistent)
                        authCookie.Expires = ticket.Expiration;

                    // S'afegeix l'etiqueta de la sessió a la resposta al client.
                    HttpContextBase httpContext = (HttpContextBase)ServiceContext.GetService(typeof(HttpContextBase));
                    httpContext.Response.Cookies.Add(authCookie);

                    // Es retorna l'usuari validat.
                    return foundUser;
                }
                // Si l'usuari no és vàlid es retorna la sessió de l'usuari anònim.
                return AnonymousUser;
            }
            catch (Exception ex)
            {
                // Es retorna l'excepció més primitiva.
                Exception actualException = ex;
                while (actualException.InnerException != null)
                    actualException = actualException.InnerException;
                throw actualException;
            }
        }

        #endregion

        #region Method IAuthentication<LoginUser>.Logout

        /// <summary>
        /// Tancament de sessió del usuari.
        /// </summary>
        /// <returns>
        /// Usuari anònim.
        /// </returns>
        public SessionUser Logout()
        {
            // Es finalitza la sessió del usuari.
            FormsAuthentication.SignOut();
            // Es retorna la sessió del usuari anònim.
            return AnonymousUser;
        }

        #endregion

        #endregion

        #region Private methods

        #region Private Method GetDirectorateUserByName

        /// <summary>
        /// Obté l'usuari d'una direcció general mitjançant el seu nom d'usuari.
        /// </summary>
        /// <param name="directorateId">
        /// Codi de la direcció general de l'usuari.
        /// </param>
        /// <param name="username">
        /// Nom d'usuari.
        /// </param>
        /// <returns>
        /// Usuari de la direcció general amb el nom d'usuari especificat.
        /// </returns>
        private SessionUser GetDirectorateUserByName(int directorateId, string username)
        {
            // Cerca l'usuari de la direcció general amb el nom d'usuari concrets.
            User foundUser = ObjectContext.Users
                .Where(u => u.DirectorateId == directorateId)
                .FirstOrDefault(u => u.UserName == username);
            // Si l'usuari existeix es retorna l'usuari d'autentificació.
            if (foundUser != null)
                return foundUser as SessionUser;
            // Si no s'ha trobat cap usuari es retorna null.
            return null;
        }

        #endregion

        #region Private Method ValidateDirectorateUser

        /// <summary>
        /// Comprova si un usuari disposa d'una subscripció amb el nom d'usuari i contrasenya especificats
        /// en una direcció general concreta.
        /// </summary>
        /// <param name="directorateId">
        /// Codi de la direcció general de l'usuari.
        /// </param>
        /// <param name="username">
        /// Nom d'usuari.
        /// </param>
        /// <param name="password">
        /// Funció lambda que retorna la contrasenya del usuari.
        /// </param>
        /// <param name="foundLoginUser">
        /// Retorna l'usuari validat.
        /// </param>
        /// <returns>
        /// Vertader si l'usuari i contrasenya són vàlids i fals si no ho són.
        /// </returns>
        private bool ValidateDirectorateUser(int directorateId, string username, Func<string> password, out SessionUser foundLoginUser)
        {
            // Cerca l'usuari de la direcció general mitjançant el seu nom d'usuari.
            foundLoginUser = GetDirectorateUserByName(directorateId, username);

            // Si existeix l'usuari es comprova la seva contrasenya.
            if (foundLoginUser != null)
                // Es recorren totes les subscripcions per comprovar si l'usuari s'ha validat correctament:
                foreach (Sigcat.Server.Model.Membership member in foundLoginUser.Memberships)
                {
                    // Es genera la clau hash mitjançant la contrasenya de login
                    // i la paraula de Salt de la subscripció:
                    string passwordHash = HashHelper.ComputeSaltedHash(password, member.PasswordSalt);
                    // Es comprova si la clau hash generada és la mateixa que la de la subscripció:
                    if (string.Equals(passwordHash, member.PasswordHash, StringComparison.Ordinal))
                        // Si les claus hash coincideixen l'usuari s'ha validat correctament.
                        return true;
                }
            // Si no hi ha cap subscripció que coincideixi
            // amb l'usuari i contrasenya especificats l'usuari no és vàlid.
            return false;
        }

        #endregion

        #endregion

    }

}
