﻿using System;

namespace fgxavi.Mvvm
{

    #region delegate AcceptPendingChangeHandler

    /// <summary>
    /// Delegat que encapsula el mètode del event <see cref="PendingChange"/>
    /// de la interfície <see cref="IAcceptPendingChange"/>.
    /// </summary>
    /// <param name="sender">
    /// Objecte que produeix l'event.
    /// </param>
    /// <param name="e">
    /// Informació del canvi de valor d'una propietat del objecte.
    /// Aquesta informació es retorna a l'emissor.
    /// </param>
    public delegate void AcceptPendingChangeHandler(object sender, out AcceptPendingChangeEventArgs e);

    #endregion

    #region Interface IAcceptPendingChange

    /// <summary>
    /// Interfície que permet implementar
    /// l'acceptació de canvis en una propietat.
    /// </summary>
    public interface IAcceptPendingChange
    {

        /// <summary>
        /// Event que desencadenarà l'objecte que implementi la interfície
        /// <see cref="IAcceptPendingChange"/> quan sol·liciti modificar 
        /// el valor d'una propietat.
        /// </summary>
        event AcceptPendingChangeHandler AcceptPendingChange;

    }

    #endregion

    #region Class AcceptPendingChangeEventArgs

    /// <summary>
    /// Informació del canvi de valor d'una propietat
    /// que es facilita al controlador d'un event de tipus
    /// <see cref="AcceptPendingChangeHandler"/> que permet
    /// decidir al Controlador si es produiran o no el canvi de valor.
    /// </summary>
    /// <remarks>
    /// Per a que el controlador del event pugui determinar
    /// l'aceptació del canvi de valor la classe
    /// <see cref="AcceptPendingChangeEventArgs"/> ha de
    /// estar declarada com a paràmetre de sortida (out).
    /// </remarks>
    public class AcceptPendingChangeEventArgs : EventArgs
    {

        #region Constructor

        /// <summary>
        /// Constructor de la classe <see cref="AcceptPendingChangeEventArgs"/>.
        /// </summary>
        /// <param name="propertyName">Nom de la propietat que sol·licita canviar de valor.</param>
        /// <param name="newValue">Nou valor que la propietat sol·licita adquirir.</param>
        public AcceptPendingChangeEventArgs(string propertyName, object newValue)
        {
            PropertyName = propertyName;
            NewValue = newValue;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Nom de la propietat que sol·licita canviar de valor.
        /// </summary>
        public string PropertyName { get; private set; }

        /// <summary>
        /// Nou valor que la propietat sol·licita adquirir.
        /// </summary>
        public object NewValue { get; private set; }

        /// <summary>
        /// Determina si es produirà o no el canvi de valor
        /// mitjançant el Controlador del event.
        /// </summary>
        public bool CancelPendingChange { get; set; }

        #endregion

    }

    #endregion

}
