﻿using Newtonsoft.Json;
using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.DataAnnotations;
using System.Data;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;

namespace fgxavi.Mvvm
{

    #region delegate ModelEditEventHandler

    /// <summary>
    /// Delegat que encapsula els mètodes dels events <see cref="ModelBase.ModelBeginEdit"/>
    /// i <see cref="ModelBase.ModelEndEdit"/> d'un model de dades <see cref="ModelBase"/>.
    /// </summary>
    /// <param name="sender">
    /// Instàcia del model de dades <see cref="ModelBase"/>
    /// que produeix l'event.
    /// </param>
    public delegate void ModelEditEventHandler(ModelBase sender);

    #endregion

    /// <summary>
    /// Enllaç de dades d'un model de dades amb totes les implementacions necessàries
    /// per alliberar la capa de dades de la capa de presentació.
    /// 
    /// Implementa la interfície <see cref="INotifyPropertyChanged"/>
    /// per notificar cada cop que una propietat ha canviat de valor.
    /// 
    /// Implementa la interfície <see cref="INotifyPropertyChanging"/>
    /// per notificar cada cop que una propietat canviarà de valor.
    /// 
    /// Implementa la interfície <see cref="ISelectable"/>
    /// per utilitzar el model de dades com un element seleccionable.
    /// 
    /// Implementa la interfície <see cref="IChangeTracking"/> per aplicar els canvis
    /// produits al model de dades i notificar-ho a la capa de negoci.
    /// 
    /// Implementa la interfície <see cref="IRevertibleChangeTracking"/>
    /// per revertir els canvis produits al model de dades.
    /// 
    /// Implementa la interfície <see cref="IEditableObject"/> per fer
    /// editable el model de dades i alliberar la capa de dades de la capa de negoci.
    /// 
    /// Implementa la interfície <see cref="ICloneable"/> per 
    /// donar suport a la interfície <see cref="IEditableObject"/>.
    /// 
    /// Implementa la interfície <see cref="INotifyDataErrorInfo"/>
    /// per notificar cada cop que un canvi en una propietat produeix
    /// un error de validació i poder informar a la capa de presentació.
    /// 
    /// Implementa la interfície <see cref="IAcceptPendingChange"/>
    /// per autoritzar els canvis de valor de les propietats.
    /// </summary>
    public class ModelBase
        : INotifyPropertyChanged, INotifyPropertyChanging, ISelectable,
        IChangeTracking, IRevertibleChangeTracking, IEditableObject, ICloneable,
        INotifyDataErrorInfo, IAcceptPendingChange
    {

        #region Member _lock

        /// <summary>
        /// Objecte privat de control que s'utilitza per bloquejar
        /// el model de dades de possibles processos concurrents.
        /// </summary>
        private readonly object _lock = new object();

        #endregion

        #region ISelectable implementation

        #region Event ISelectable.SelectionChanged

        /// <summary>
        /// Event que es produeix quan es modifica la selecció del objecte.
        /// </summary>
        public event SelectionChangedHandler SelectionChanged;

        #endregion

        #region Controller OnSelectionChanged

        /// <summary>
        /// Controlador que desencadena l'event <see cref="SelectionChanged"/>.
        /// </summary>
        /// <param name="sender">
        /// Objecte seleccionable <see cref="ISelectable"/> que produeix l'event.
        /// </param>
        /// <param name="selection">
        /// Informa al controlador si s'ha seleccionat l'objecte o si s'ha deseleccionat.
        /// </param>
        protected void OnSelectionChanged(ISelectable sender, bool selection)
        {
            if (SelectionChanged != null)
                SelectionChanged(sender, selection);
            // Actualitza les propietats relacionades.
            RaisePropertyChanged(() => Selected);
        }

        #endregion

        #region Property ISelectable.Selected

        /// <summary>
        /// Membre privat de la propietat <see cref="Selected" />.
        /// </summary>
        private bool _Selected = false;

        /// <summary>
        /// Estat de selecció del objecte.
        /// És vertader si l'objecte està seleccionat i fals si no és així.
        /// Estableix i obté la propietat <see cref="Selected" />.
        /// </summary>
        [JsonIgnore]
        public bool Selected
        {
            get { return _Selected; }
            set
            {
                if (_Selected != value)
                {
                    // S'assigna el nou valor.
                    _Selected = value;
                    // Crida el controlador de la interfície INotifyPropertyChanged.
                    RaisePropertyChanged();
                    // Crida el controlador de la interfície ISelectable.
                    OnSelectionChanged(this, _Selected);
                }
            }
        }

        #endregion

        #endregion

        #region INotifyPropertyChanged implementation

        #region Event INotifyPropertyChanged.PropertyChanged

        /// <summary>
        /// Event que es produeix quan es modifica una propietat.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        #region Controller RaisePropertyChanged([CallerMemberName]string propertyName = null)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanged"/>
        /// un cop s'ha modificat un camp o una propietat.
        /// </summary>
        /// <param name="propertyName">
        /// Nom de la propietat que s'ha modificat.
        /// </param>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        protected virtual void RaisePropertyChanged([CallerMemberName]string propertyName = null)
        {
            // Si hi ha un Controlador enllaçat a l'event PropertyChanged l'executa.
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

        #region Controller RaisePropertyChanged<T>(Expression<Func<T>> expression)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanged"/>
        /// un cop s'ha modificat un camp o una propietat.
        /// </summary>
        /// <typeparam name="T">
        /// Tipus retornat per l'expressió lambda del paràmetre <paramref name="expression"/>.
        /// </typeparam>
        /// <param name="expression">
        /// Expressió lambda que ha de retornar el camp o la propietat que s'ha modificat.
        /// </param>
        /// <remarks>
        /// Si l'expressió lambda no retorna un camp o una propietat es produeixen excepcions.
        /// </remarks>
        protected virtual void RaisePropertyChanged<T>(Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de la implementació INotifyPropertyChanged.
            RaisePropertyChanged(body.Member.Name);
        }

        #endregion

        #endregion

        #region INotifyPropertyChanging implementation

        #region Event INotifyPropertyChanging.PropertyChanging

        /// <summary>
        /// Event que es produeix abans de modificar-se una propietat.
        /// </summary>
        public event PropertyChangingEventHandler PropertyChanging;

        #endregion

        #region Controller RaisePropertyChanging([CallerMemberName]string propertyName = null)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanging"/>
        /// abans de modificar-se una propietat.
        /// </summary>
        /// <param name="propertyName">
        /// Nom de la propietat que es modificarà.
        /// </param>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        protected virtual void RaisePropertyChanging([CallerMemberName]string propertyName = null)
        {
            // Si hi ha un Controlador enllaçat a l'event PropertyChanging l'executa.
            if (PropertyChanging != null)
                PropertyChanging(this, new PropertyChangingEventArgs(propertyName));
        }

        #endregion

        #region Controller RaisePropertyChanging<T>(Expression<Func<T>> expression)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanging"/>
        /// abans de modificar-se una propietat.
        /// </summary>
        /// <typeparam name="T">
        /// Tipus retornat per l'expressió lambda del paràmetre <paramref name="expression"/>.
        /// </typeparam>
        /// <param name="expression">
        /// Expressió lambda que ha de retornar el camp o la propietat que abans de modificar-se.
        /// </param>
        /// <remarks>
        /// Si l'expressió lambda no retorna un camp o una propietat es produeixen excepcions.
        /// </remarks>
        protected virtual void RaisePropertyChanging<T>(Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de la implementació INotifyPropertyChanging.
            RaisePropertyChanging(body.Member.Name);
        }

        #endregion

        #endregion

        #region IChangeTracking implementation

        #region Events AcceptChanges

        /// <summary>
        /// Event que es produeix quan s'accepten els canvis del model de dades.
        /// </summary>
        public event SendChangesEventHandler<ModelBase> SendAcceptChanges;

        /// <summary>
        /// Event que es produeix quan s'accepten els canvis del model de dades.
        /// </summary>
        public event RecieveChangesEventHandler<ModelBase> RecieveAcceptChanges;

        #endregion

        #region Controllers AcceptChanges

        #region Controller OnSendAcceptChanges

        /// <summary>
        /// Controlador que desencadena l'event <see cref="SendAcceptChanges"/>
        /// de la col·lecció de models de dades.
        /// </summary>
        /// <param name="sender">Model de dades que produeix l'event.</param>
        protected void OnSendAcceptChanges(ModelBase sender)
        {
            if (SendAcceptChanges != null)
                SendAcceptChanges(sender);
        }

        #endregion

        #region Controller OnRecieveAcceptChanges

        /// <summary>
        /// Controlador que desencadena l'event <see cref="RecieveAcceptChanges"/>
        /// de la col·lecció de models de dades.
        /// </summary>
        /// <param name="sender">Model de dades que produeix l'event.</param>
        protected void OnRecieveAcceptChanges(ModelBase sender, ChangeResultEventArgs<ModelBase> args)
        {
            if (RecieveAcceptChanges != null)
                RecieveAcceptChanges(sender, args);
        }

        #endregion

        #endregion

        #region Property State

        /// <summary>
        /// Membre privat de la propietat <see cref="State" />.
        /// </summary>
        private DataRowState _State = DataRowState.Unchanged;

        /// <summary>
        /// Estat del model de dades definit per l'enumeració <see cref="DataRowState"/>.
        /// El valor UNCHANGED informa que el model de dades no s'ha modificat.
        /// El valor ADDED informa que el model de dades és nou.
        /// El valor MODIFIED informa que el model de dades s'ha modificat.
        /// El valor DELETED informa que el model de dades s'ha esborrat.
        /// El valor DETACHED informa que el model de dades s'ha creat
        /// però no ha finalitzat la seva edició.
        /// Estableix i obté la propietat <see cref="State" />.
        /// Fa ús del mètode <see cref="SetField"/> per establir la propietat.
        /// </summary>
        [JsonIgnore]
        public DataRowState State
        {
            get { return _State; }
            set
            {
                if (_State != value)
                {
                    // S'assigna el nou valor.
                    _State = value;
                    // S'informa que el model de dades ha canviat.
                    IsChanged = true;
                    // Crida el controlador de la interfície INotifyPropertyChanged.
                    RaisePropertyChanged();
                }
            }
        }

        #endregion

        #region Property IChangeTracking.IsChanged

        /// <summary>
        /// Membre privat de la propietat <see cref="IsChanged" />.
        /// </summary>
        private bool _IsChanged = false;

        /// <summary>
        /// Estat de canvi del model de dades.
        /// És vertader si el model de dades s'ha modificat i fals si no és així.
        /// Estableix i obté la propietat <see cref="IsChanged" />.
        /// Fa ús del mètode <see cref="SetField"/> per establir la propietat.
        /// </summary>
        [JsonIgnore]
        public bool IsChanged
        {
            get { return _IsChanged; }
            protected set
            {
                if (_IsChanged != value)
                {
                    // S'assigna el nou valor.
                    _IsChanged = value;
                    // Crida el Controlador de la interfície INotifyPropertyChanged.
                    RaisePropertyChanged();
                }
            }
        }

        #endregion

        #region Method Backup

        /// <summary>
        /// Emmagatzema una còpia del model de dades actual.
        /// </summary>
        public void Backup()
        {
            // Emmagatzema una còpia del model de dades actual.
            OriginalBackup = this.Clone() as ModelBase;
        }

        #endregion

        #region Method IChangeTracking.AcceptChanges

        /// <summary>
        /// Desencadena l'event <see cref="SendAcceptChanges"/> per sol·licitar
        /// a la capa de la lògica de negoci que apliqui els canvis del model de dades.
        /// </summary>
        public void AcceptChanges()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Si el model de dades no ha patit cap modificació surt.
                if (State == DataRowState.Unchanged
                    || State == DataRowState.Detached)
                    return;
                // Desencadena l'event de canvis.
                OnSendAcceptChanges(this);
            }
        }

        public void AcceptChangesResult(ChangeResultEventArgs<ModelBase> args)
        {
            if (args.Result.HasError
                || args.Result.EntityResult == null
                || args.Result.EntityResult.Entity == null)
                throw new Exception(args.Result.ErrorResult.ErrorMessage);
            if (args.Result.EntityResult.ValidationErrors != null)
                foreach (ChangeValidationError error in args.Result.EntityResult.ValidationErrors)
                {
                    string[] errorMessages = new string[] { error.Message };
                    foreach (string propertyName in error.SourceMemberNames)
                    {
                        // S'afegeix el missatge d'error al diccionari.
                        _errors.TryAdd(propertyName, errorMessages.ToList());
                        // Es desencadena l'event d'error d'aquest membre per a que el Controlador
                        // comprovi GetErrors(propertyName) amb els errors actuals.
                        RaiseErrorsChanged(propertyName);
                    }
                }
            else
            {
                if (State != DataRowState.Deleted)
                {
                    // Actualitza la informació del servidor.
                    this.Update(args.Result.EntityResult.Entity);
                    // Es modifica l'estat del model de dades.
                    State = DataRowState.Unchanged;
                    // Informa que el model de dades no conté modificacions.
                    IsChanged = false;
                    // Emmagatzema una còpia del model de dades.
                    Backup();
                }
                OnRecieveAcceptChanges(this as ModelBase, args);
            }
        }

        #endregion

        #endregion

        #region IRevertibleChangeTracking implementation

        #region Events RejectChanges

        /// <summary>
        /// Event que es produeix quan es restauren els canvis del model de dades.
        /// </summary>
        public event SendChangesEventHandler<ModelBase> SendRejectChanges;

        #endregion

        #region Controllers RejectChanges

        #region Controller OnModelRejectChanges

        /// <summary>
        /// Controlador que desencadena l'event <see cref="SendRejectChanges"/>
        /// de la col·lecció de models de dades.
        /// </summary>
        /// <param name="sender">Model de dades que produeix l'event.</param>
        protected void OnSendRejectChanges(ModelBase sender)
        {
            if (SendRejectChanges != null)
                SendRejectChanges(sender);
        }

        #endregion

        #endregion

        #region Property OriginalBackup

        /// <summary>
        /// Membre privat de la propietat <see cref="OriginalBackup" />.
        /// </summary>
        private ModelBase _OriginalBackup = null;

        /// <summary>
        /// Còpia del model de dades. S'emmagatzema el model de dades original
        /// obtingut inicialment mitjançant la lògica de negoci.
        /// Estableix i obté la propietat <see cref="OriginalBackup" />.
        /// Fa ús del mètode <see cref="SetField"/> per establir la propietat.
        /// </summary>
        /// <remarks>
        /// Pels models de dades creats amb estat ADDED no s'emmagatzemarà res.
        /// </remarks>
        [JsonIgnore]
        public ModelBase OriginalBackup
        {
            get { return _OriginalBackup; }
            protected set
            {
                // S'assigna el nou valor.
                _OriginalBackup = value;
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Method Restore

        /// <summary>
        /// Restaura la còpia del model de dades actual.
        /// </summary>
        public void Restore()
        {
            // Restaura el model de dades al seu estat original.
            Update(OriginalBackup);
        }

        #endregion

        #region Method IRevertibleChangeTracking.RejectChanges

        /// <summary>
        /// Retorna el model de dades al seu estat original
        /// abans de totes les modificacions.
        /// </summary>
        /// <remarks>
        /// Si el model de dades és nou aleshores no fa res.
        /// </remarks>
        public void RejectChanges()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Desencadena l'event de canvis independentment de l'estat del model.
                OnSendRejectChanges(this);
                // Si el model de dades no ha patit cap modificació surt.
                if (State == DataRowState.Unchanged)
                    return;
                // Restaura el model de dades al seu estat original.
                Restore();
                // Informa que el model de dades no conté modificacions.
                IsChanged = false;
                // Informa que el model de dades no conté cap modificació.
                State = DataRowState.Unchanged;
            }
        }

        #endregion

        #endregion

        #region IEditableObject implementation

        #region Events

        /// <summary>
        /// Event que es produeix quan es comença a editar el model de dades.
        /// </summary>
        public event ModelEditEventHandler ModelBeginEdit;

        /// <summary>
        /// Event que es produeix quan es s'acaba d'editar el model de dades.
        /// </summary>
        public event ModelEditEventHandler ModelEndEdit;

        #endregion

        #region Controllers

        #region Controller OnModelBeginEdit

        /// <summary>
        /// Controlador que desencadena l'event <see cref="ModelBeginEdit"/>.
        /// </summary>
        /// <param name="sender">Model de dades <see cref="ModelBase"/> que produeix l'event.</param>
        protected void OnModelBeginEdit(ModelBase sender)
        {
            if (ModelBeginEdit != null)
                ModelBeginEdit(sender);
        }

        #endregion

        #region Controller OnModelEndEdit

        /// <summary>
        /// Controlador que desencadena l'event <see cref="ModelEndEdit"/>.
        /// </summary>
        /// <param name="sender">
        /// Model de dades <see cref="ModelBase"/> que produeix l'event.
        /// </param>
        protected void OnModelEndEdit(ModelBase sender)
        {
            if (ModelEndEdit != null)
                ModelEndEdit(sender);
        }

        #endregion

        #endregion

        #region Properties

        #region Property EditingBackup

        /// <summary>
        /// Membre privat de la propietat <see cref="EditingBackup" />.
        /// </summary>
        private ModelBase _EditingBackup = null;

        /// <summary>
        /// Còpia del model de dades. S'emmagatzema l'estat del model de dades abans 
        /// de la seva edició per poder restaurar-lo si es cancel·la l'edició.
        /// Estableix i obté la propietat <see cref="EditingBackup" />.
        /// Fa ús del mètode <see cref="SetField"/> per establir la propietat.
        /// </summary>
        [JsonIgnore]
        public ModelBase EditingBackup
        {
            get { return _EditingBackup; }
            protected set
            {
                // S'assigna el nou valor.
                _EditingBackup = value;
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Property IsEditing

        /// <summary>
        /// Membre privat de la propietat <see cref="IsEditing" />.
        /// </summary>
        private bool _IsEditing = false;

        /// <summary>
        /// Estat de l'edició del model de dades.
        /// És vertader si el model de dades s'està editant i fals si no és així.
        /// Estableix i obté la propietat <see cref="IsEditing" />.
        /// Fa ús del mètode <see cref="SetField"/> per establir la propietat.
        /// </summary>
        [JsonIgnore]
        public bool IsEditing
        {
            get { return _IsEditing; }
            protected set
            {
                if (_IsEditing != value)
                {
                    // S'assigna el nou valor.
                    _IsEditing = value;
                    // Crida el Controlador de la interfície INotifyPropertyChanged.
                    RaisePropertyChanged();
                }
            }
        }

        #endregion

        #endregion

        #region IEditableObject.BeginEdit

        /// <summary>
        /// Desencadena l'event <see cref="ModelBeginEdit"/> per començar l'edició.
        /// </summary>
        public void BeginEdit()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Si el model de dades ja s'està editant surt.
                if (IsEditing)
                    return;
                // Posa en estat d'edició el model de dades.
                IsEditing = true;
                // Realitza una còpia del estat del model de dades per poder revertir els canvis.
                EditingBackup = this.Clone() as ModelBase;
                // Desencadena l'event ModelBeginEdit.
                OnModelBeginEdit(this);
            }
        }

        #endregion

        #region IEditableObject.CancelEdit

        /// <summary>
        /// Retorna el model de dades al seu estat anterior a l'edició.
        /// </summary>
        public void CancelEdit()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Si el model de dades no s'està editant surt.
                if (!IsEditing)
                    return;
                // Surt de l'estat d'edició del model de dades.
                IsEditing = false;
                // Restaura l'estat del model de dades a abans de l'edició.
                Update(EditingBackup);
            }
        }

        #endregion

        #region IEditableObject.EndEdit

        /// <summary>
        /// Desencadena l'event <see cref="ModelEndEdit"/> per finalitzar l'edició.
        /// </summary>
        public void EndEdit()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Si el model de dades no s'està editant surt.
                if (!IsEditing)
                    return;
                // Surt de l'estat d'edició del model de dades.
                IsEditing = false;
                // Destrueix la còpia del model de dades.
                EditingBackup = null;
                // S'informa que s'han produït modificacions.
                IsChanged = true;
                // Canvia l'estat del model de dades.
                switch (State)
                {
                    case DataRowState.Detached:
                        State = DataRowState.Added;
                        break;
                    case DataRowState.Unchanged:
                        State = DataRowState.Modified;
                        break;
                    case DataRowState.Added:
                    case DataRowState.Deleted:
                    case DataRowState.Modified:
                    default:
                        break;
                }
                // Desencadena l'event ModelEndEdit.
                OnModelEndEdit(this);
            }
        }

        #endregion

        #region IClonable.Clone

        /// <summary>
        /// Crea una nova instància còpia de l'actual 
        /// replicant tots els membres de la classe 
        /// implementant la interfície <see cref="IClonable"/>.
        /// </summary>
        /// <returns>Objecte còpia de l'actual.</returns>
        public object Clone()
        {
            return this.MemberwiseClone();
        }

        #endregion

        #region Private Method Update

        /// <summary>
        /// Replica el model de dades <paramref name="obj"/> en la instància actual
        /// actualitzant totes les dades fent ús de <see cref="Reflexion"/>.
        /// </summary>
        /// <param name="obj">Objecte per replicar.</param>
        /// <remarks>
        /// Només còpia els camps declarats en els tipus que hereten
        /// de <see cref="ModelBase"/>.
        /// </remarks>
        public void Update(ModelBase obj)
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Si el model de dades no està instanciat surt.
                if (obj == null)
                    return;
                // Recorre l'herència del model de dades a partir del seu tipus
                // fins a la classe actual no inclosa.
                Type currentType = obj.GetType();
                Type endType = typeof(ModelBase);
                while (!currentType.Equals(endType))
                {
                    // Utilitza Reflexion per obtenir tots els camps de la instància.
                    FieldInfo[] fields = currentType.GetFields(
                        BindingFlags.NonPublic | BindingFlags.Public |
                        BindingFlags.Instance);
                    // Utilitza Reflexion per copiar un a un tots els camps de la instància.
                    foreach (FieldInfo fi in fields)
                        // Només còpia els camps si estan declarats al tipus analitzat.
                        if (fi.DeclaringType.Equals(currentType))
                            fi.SetValue(this, fi.GetValue(obj));
                    // Utilitza Reflexion per obtenir totes les propietats de la instància.
                    PropertyInfo[] props = currentType.GetProperties(
                        BindingFlags.NonPublic | BindingFlags.Public |
                        BindingFlags.Instance);
                    // Utilitza Reflexion per informar que els valors de les propietats han canviat.
                    foreach (PropertyInfo pi in props)
                        // Només informa si estan declarats al tipus analitzat.
                        if (pi.DeclaringType.Equals(currentType))
                            RaisePropertyChanged(pi.Name);
                    // Obté el l'antecessor del tipus analitzat.
                    currentType = currentType.BaseType;
                }
            }
        }

        #endregion

        #endregion

        #region INotifyDataErrorInfo implementation

        #region Event INotifyDataErrorInfo.ErrorsChanged

        /// <summary>
        /// Event que es produeix quan el valor
        /// que es vol assignar a una propietat conté un error.
        /// </summary>
        public event EventHandler<DataErrorsChangedEventArgs> ErrorsChanged;

        #endregion

        #region Member _errors

        /// <summary>
        /// Registre dels errors de cadascun dels camps o de les propietats de la classe.
        /// S'emmagatzema la informació en un diccionari segons el nom del camp o de la propietat.
        /// </summary>
        protected ConcurrentDictionary<string, List<string>> _errors =
            new ConcurrentDictionary<string, List<string>>();

        #endregion

        #region Method INotifyDataErrorInfo.GetErrors

        /// <summary>
        /// Obté una enumeració de tots els errors d'un camp o d'una propietat,
        /// s'implementa de la interfície <see cref="INotifyDataErrorInfo"/>.
        /// </summary>
        /// <param name="propertyName">
        /// Nom del camp o de la propietat del qual es cerquen errors.
        /// </param>
        /// <returns>
        /// Enumeració de tots els errors cercats del camp o de la propietat.
        /// </returns>
        public IEnumerable GetErrors(string propertyName)
        {
            // S'obtenen tots els errors d'una clau del diccionari.
            List<string> propertyErrors;
            _errors.TryGetValue(propertyName, out propertyErrors);
            return propertyErrors;
        }

        #endregion

        #region Property INotifyDataErrorInfo.HasErrors

        /// <summary>
        /// Propietat de lectura que indica si s'han enregistrat errors en tota la classe.
        /// </summary>
        [JsonIgnore]
        public bool HasErrors
        {
            get { return _errors.Any(kv => kv.Value != null && kv.Value.Count > 0); }
        }

        #endregion

        #region Controller RaiseErrorsChanged([CallerMemberName]string propertyName = null)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="ErrorsChanged"/>
        /// quan es produeixen errors en el canvi de valor d'un camp o d'una propietat.
        /// </summary>
        /// <param name="propertyName">
        /// Nom de la propietat que s'ha modificat.
        /// </param>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        public void RaiseErrorsChanged([CallerMemberName]string propertyName = null)
        {
            // Si hi ha un Controlador enllaçat a l'event ErrorsChanged l'executa.
            if (ErrorsChanged != null)
                ErrorsChanged(this, new DataErrorsChangedEventArgs(propertyName));
        }

        #endregion

        #region Controller RaiseErrorsChanged<T>(Expression<Func<T>> expression)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="ErrorsChanged"/>
        /// quan es produeixen errors en el canvi de valor d'un camp o d'una propietat.
        /// </summary>
        /// <typeparam name="T">
        /// Tipus retornat per l'expressió lambda del paràmetre <paramref name="expression"/>.
        /// </typeparam>
        /// <param name="expression">
        /// Expressió lambda que ha de retornar el camp o la propietat que ha produït error.
        /// </param>
        /// <remarks>
        /// Si l'expressió lambda no retorna un camp o una propietat es produeixen excepcions.
        /// </remarks>
        protected virtual void RaiseErrorsChanged<T>(Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de la implementació INotifyErrorsChanged.
            RaiseErrorsChanged(body.Member.Name);
        }

        #endregion

        #region Method ValidateAsync

        /// <summary>
        /// Valida els errors de tots els camps i propietats de la classe
        /// de forma asíncrona.
        /// </summary>
        /// <returns>
        /// Tasca <see cref="Task"/> executant la validació asíncrona.
        /// </returns>
        public Task ValidateAsync()
        {
            return Task.Run(() => Validate());
        }

        #endregion

        #region Method Validate

        /// <summary>
        /// Valida els errors de tots els camps i propietats de la classe.
        /// </summary>
        /// <remarks>
        /// L'algoritme està bloquejat contra l'accés concurrent.
        /// </remarks>
        public void Validate()
        {
            // Es protegeix l'algoritme amb la sentència lock i el camp de control _lock
            // per a que cap procés l'executi mentre algun altre procés l'està executant.
            lock (_lock)
            {

                #region S'obtenen tots els ValidationResults del model de dades sencer

                // Es crea un contenidor de les validacions negatives de les propietats.
                var validationResults = new List<ValidationResult>();
                // Es declara el context de validació per a que el Controlador Validator
                // conegui les regles de validació que ha d'utilitzar per validar el model de dades.
                ValidationContext validationContext;

                // Recorre tota la herència de classes des de la instància actual 
                // fins a aquesta classe cercant totes les classes Metadata 
                // per aplicar les validacions d'errors sobre cadascuna d'elles.
                Type currentType = this.GetType();
                while (!currentType.Equals(typeof(ModelBase)))
                {
                    // Obté tots els tipus de les classes informades mitjançant MetadataTypeAttribute
                    // sobre la declaració del tipus de la classe analitzada.
                    var MetadataClassTypes = from attr in Attribute.GetCustomAttributes(currentType, true)
                                             where attr is MetadataTypeAttribute
                                             select (attr as MetadataTypeAttribute).MetadataClassType;
                    // Per cada tipus de classe de metadates informat es crea una instància
                    // i es valida el model de dades.
                    foreach (Type metadataClassType in MetadataClassTypes)
                    {
                        // Es capturen les possibles excepcions.
                        try
                        {
                            // Es crea una instància de la classe de les metadates sense constructor.
                            var metadataInstance = Activator.CreateInstance(metadataClassType);
                            // Es crea el context de validació per a que el Controlador Validator
                            // sàpiga que ha de validar les metadates d'aquest model de dades.
                            validationContext = new ValidationContext(metadataInstance);
                            // Es valida la propietat utilitzant el Controlador Validator
                            // que obté tots els ValidationResults associats a la propietat
                            // i retorna només aquells que han sigut negatius.
                            Validator.TryValidateObject(this, validationContext, validationResults, true);
                        }
                        catch (Exception) { }
                    }
                    // Passa a analitzar el tipus de la classe pare.
                    currentType = currentType.BaseType;
                }

                // S'assigna aquesta instància com a context de validació pel Controlador Validator.
                validationContext = new ValidationContext(this);
                // S'obtenen els resultats de les validacions d'errors de la classe actual.
                Validator.TryValidateObject(this, validationContext, validationResults, true);

                #endregion

                #region S'actualitzen els errors del model de dades al diccionari d'errors

                #region S'esborra del diccionari tots els errors que ja no es produeixen

                // Es comproven tots els errors enregistrats 
                // per eliminar els errors que ja no es produeixen,
                // per fer això es recorren tots els membres enregistrats amb errors.
                foreach (var kv in _errors.ToList())
                    // Només si totes les validacions no detecten error en el membre
                    // s'esborra el membre amb error.
                    if (validationResults.All(r => r.MemberNames.All(m => m != kv.Key)))
                    {
                        // S'elimina el membre del diccionari contenidor d'errors.
                        List<string> outLi;
                        _errors.TryRemove(kv.Key, out outLi);
                        // Es desencadena l'event d'error d'aquest membre per a que el Controlador
                        // comprovi GetErrors(propertyName) i sàpiga que ja no conté errors.
                        RaiseErrorsChanged(kv.Key);
                    }

                #endregion

                #region S'afegeixen al diccionari tots els errors produits.

                // S'agrupen tots els errors detectats segons el nom del seu membre.
                var members = from vr in validationResults
                              from mn in vr.MemberNames
                              group vr by mn into k
                              select k;
                // Es recorren tots els membres amb errors detectats.
                foreach (var member in members)
                {
                    // S'obtenen tots els missatges d'error detectats en aquest membre.
                    var messages = member.Select(r => r.ErrorMessage).ToList();
                    // Si el membre existeix al diccionari s'esborra.
                    if (_errors.ContainsKey(member.Key))
                    {
                        List<string> outLi;
                        _errors.TryRemove(member.Key, out outLi);
                    }
                    // S'afegeix el membre al diccionari amb els seus missatges d'error detectats.
                    _errors.TryAdd(member.Key, messages);
                    // Es desencadena l'event d'error d'aquest membre per a que el Controlador
                    // comprovi GetErrors(propertyName) amb els errors actuals.
                    RaiseErrorsChanged(member.Key);
                }

                #endregion

                #endregion

            }
        }

        #endregion

        #region Method ValidateMemberValue

        /// <summary>
        /// Valida el valor del camp o la propietat especificat.
        /// </summary>
        /// <param name="newValue">
        /// Valor de la propietat a validar.
        /// </param>
        /// <param name="propertyName">
        /// Nom de la propietat a validar el valor.
        /// </param>
        /// <returns>
        /// Vertader si el valor de la propietat no conté errors i fals si en conté.
        /// </returns>
        /// <remarks>
        /// L'algoritme està bloquejat contra l'accés concurrent.
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        public bool ValidateMemberValue(object newValue, [CallerMemberName]string propertyName = null)
        {
            // Es protegeix l'algoritme amb la sentència lock i el camp de control _lock
            // per a que cap procés l'executi mentre algun altre procés l'està executant.
            lock (_lock)
            {

                #region S'obtenen tots els ValidationResults de la propietat validada

                // Es crea un contenidor de les validacions negatives de la propietat.
                var validationResults = new List<ValidationResult>();
                // Es declara el context de validació per a que el Controlador Validator
                // conegui les regles de validació que ha d'utilitzar per validar la propietat.
                ValidationContext validationContext;

                // Recorre tota la herència de classes des de la instància actual 
                // fins a aquesta classe cercant totes les classes Metadata 
                // per aplicar les validacions d'errors sobre cadascuna d'elles.
                Type currentType = this.GetType();
                while (!currentType.Equals(typeof(ModelBase)))
                {
                    // Obté tots els tipus de les classes informades mitjançant MetadataTypeAttribute
                    // sobre la declaració del tipus de la classe analitzada.
                    var MetadataClassTypes = from attr in Attribute.GetCustomAttributes(currentType, true)
                                             where attr is MetadataTypeAttribute
                                             select (attr as MetadataTypeAttribute).MetadataClassType;
                    // Per cada tipus de classe de metadates informat es crea una instància
                    // i es valida la propietat.
                    foreach (Type metadataClassType in MetadataClassTypes)
                    {
                        // Es capturen les possibles excepcions.
                        try
                        {
                            // Es crea una instància de la classe de les metadates sense constructor.
                            var metadataInstance = Activator.CreateInstance(metadataClassType);
                            // Es crea el context de validació per a que el Controlador Validator
                            // sàpiga que ha de validar la propietat propertyName d'aquest objecte.
                            validationContext = new ValidationContext(metadataInstance)
                            {
                                MemberName = propertyName
                            };
                            // Es valida la propietat utilitzant el Controlador Validator
                            // que obté tots els ValidationResults associats a la propietat
                            // i retorna només aquells que han sigut negatius.
                            Validator.TryValidateProperty(newValue, validationContext, validationResults);
                        }
                        catch (Exception) { }
                    }
                    // Passa a analitzar el tipus de la classe pare.
                    currentType = currentType.BaseType;
                }

                // S'assigna aquesta instància com a context de validació pel Controlador Validator.
                validationContext = new ValidationContext(this);
                // S'obtenen els resultats de les validacions d'errors de la classe actual.
                Validator.TryValidateObject(this, validationContext, validationResults, true);

                #endregion

                #region S'actualitzen els errors de la propietat al diccionari d'errors

                // S'actualitza el diccionari d'errors segons els errors de validació detectats.
                if (validationResults.Count == 0)
                {
                    // Si la propietat és vàlida es verifica que al diccionari d'errors
                    // no contigui errors en aquesta propietat.
                    if (_errors.Keys.Contains(propertyName))
                    {
                        // Si conté errors s'eliminen.
                        var removedErrors = new List<string>();
                        _errors.TryRemove(propertyName, out removedErrors);
                        // Es desencadena l'event d'error d'aquest membre per a que el Controlador
                        // comprovi GetErrors(propertyName) i sàpiga que ja no conté errors.
                        RaiseErrorsChanged(propertyName);
                    }
                    // Informa que la propietat no conté errors i que per tant és positiva.
                    return true;
                }
                else
                {
                    // Si la propietat no és vàlida s'afegeixen o es sobrescriuen
                    // els errors detectats actualment.
                    var messages = (from vr in validationResults
                                    select vr.ErrorMessage).ToList();
                    _errors.TryAdd(propertyName, messages);
                    // Es desencadena l'event d'error d'aquest membre per a que el Controlador
                    // comprovi GetErrors(propertyName) amb els errors actuals.
                    RaiseErrorsChanged(propertyName);
                    // Informa que la propietat conté errors i que per tant és negativa.
                    return false;
                }

                #endregion

            }
        }

        #endregion

        //#region Method SetError

        ///// <summary>
        ///// Reporta un error des de fora del model de dades especificant a quines propietats afecta.
        ///// </summary>
        ///// <param name="errorMessage">
        ///// Missatge d'error informat.
        ///// </param>
        ///// <param name="propertyNames">
        ///// Nom de les propietats del qual s'informen errors.
        ///// </param>
        //public void SetError(string errorMessage, string[] propertyNames)
        //{
        //    string[] errorMessages = new string[] { errorMessage };
        //    foreach (string propertyName in propertyNames)
        //    {
        //        // S'afegeix el missatge d'error al diccionari.
        //        _errors.TryAdd(propertyName, errorMessages.ToList());
        //        // Es desencadena l'event d'error d'aquest membre per a que el Controlador
        //        // comprovi GetErrors(propertyName) amb els errors actuals.
        //        RaiseErrorsChanged(propertyName);
        //    }
        //}

        //#endregion

        #endregion

        #region IAcceptPendingChange implementation

        #region Event IAcceptPendingChange.PendingChange

        /// <summary>
        /// Event que es produeix quan es sol·liciti modificar
        /// un valor d'una propietat del model de dades.
        /// </summary>
        public event AcceptPendingChangeHandler AcceptPendingChange;

        #endregion

        #region Controller RaiseAcceptPendingChange(object newValue, [CallerMemberName]string propertyName = null)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="AcceptPendingChange"/>
        /// per determinar si es produirà o no el canvi de valor d'una propietat.
        /// </summary>
        /// <param name="newValue">
        /// Nou valor que la propietat sol·licita adquirir.
        /// </param>
        /// <param name="propertyName">
        /// Nom de la propietat que sol·licita canviar de valor.
        /// </param>
        /// <returns>
        /// Determina si es produirà o no el canvi de valor de la propietat.
        /// </returns>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        protected virtual bool RaiseAcceptPendingChange(object newValue, [CallerMemberName]string propertyName = null)
        {
            // Si hi ha un Controlador enllaçat a l'event AcceptPendingChange l'executa.
            if (AcceptPendingChange != null)
            {
                // Facilita al Controlador informació del canvi de valor sol·licitat.
                AcceptPendingChangeEventArgs e = new AcceptPendingChangeEventArgs(propertyName, newValue);
                // Desencadena l'event de la interfície IAcceptPendingChange.
                AcceptPendingChange(this, out e);
                // Retorna la resposta del controlador d'acceptació del canvi de valor.
                return !e.CancelPendingChange;
            }
            // Si no hi ha Controlador de l'event accepta el canvi de valor per defecte.
            return true;
        }

        #endregion

        #region Controller RaiseAcceptPendingChange<T>(object newValue, Expression<Func<T>> expression)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="AcceptPendingChange"/>
        /// per determinar si es produirà o no el canvi de valor d'una propietat.
        /// </summary>
        /// <typeparam name="T">
        /// Tipus retornat per l'expressió lambda del paràmetre <paramref name="expression"/>.
        /// </typeparam>
        /// <param name="newValue">
        /// Nou valor que la propietat sol·licita adquirir.
        /// </param>
        /// <param name="expression">
        /// Expressió lambda que ha de retornar el camp o la propietat que sol·licita canviar de valor.
        /// </param>
        /// <param name="propertyName">
        /// Nom de la propietat que sol·licita canviar de valor.
        /// </param>
        /// <returns>
        /// Determina si es produirà o no el canvi de valor de la propietat.
        /// </returns>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// Si l'expressió lambda no retorna un camp o una propietat es produeixen excepcions.
        /// </remarks>
        protected virtual bool RaiseAcceptPendingChange<T>(object newValue, Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de la implementació INotifyAcceptPendingChange.
            return RaiseAcceptPendingChange(newValue, body.Member.Name);
        }

        #endregion

        #endregion

        #region Helper Methods

        #region Helper Method SetField<T>(ref T field, T value, [CallerMemberName]string propertyName = null)

        /// <summary>
        /// Mètode d'ajut al programador que fa ús de totes les implementacions de la classe
        /// <see cref="ModelBase"/> per afegir-les a cada propietat del model.
        /// </summary>
        /// <typeparam name="T">Tipus del camp o la propietat que es vol modificar.</typeparam>
        /// <param name="field">Referència al camp o a la propietat que es vol modificar.</param>
        /// <param name="value">Nou valor del camp o la propietat a modificar.</param>
        /// <param name="propertyName">Nom del camp o la propietat que es vol modificar.</param>
        /// <returns>
        /// Vertader si s'ha completat amb èxit el canvi de valor del camp o de la propietat.
        /// Fals si el canvi de valor no ha passat el procés d'acceptació del nou valor.
        /// </returns>
        /// <remarks>En cas d'errors no control·lats es propaguen les excepcions.</remarks>
        protected virtual bool SetField<T>(ref T field, T value, [CallerMemberName]string propertyName = null)
        {
            // Si el valor del camp o propietat és el mateix no es modifica.
            if (EqualityComparer<T>.Default.Equals(field, value))
                // Informa que no s'ha modificat el camp o la propietat.
                return false;
            // Valida el nou valor de la propietat cercant possibles errors,
            // cridant al Controlador de la interfície INotifyDataErrorInfo
            ValidateMemberValue(value, propertyName);
            // Crida el Controlador de la interfície IAcceptPendingChange,
            // si s'accepta el canvi de valor es realitza.
            if (RaiseAcceptPendingChange(value, propertyName))
            {
                // Crida el Controlador de la interfície INotifyPropertyChanging.
                RaisePropertyChanging(propertyName);
                // S'assigna el nou valor.
                field = value;
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged(propertyName);
                // Informa que s'ha modificat el camp o la propietat.
                return true;
            }
            // Informa que no s'ha acceptat el canvi de valor del camp o la propietat.
            return false;
        }

        #endregion

        #region Helper Method SetField<T>(ref T field, T value, Expression<Func<T>> expression)

        /// <summary>
        /// Mètode d'ajut al programador que fa ús de totes les implementacions de la classe
        /// <see cref="ModelBase"/> per afegir-les a cada propietat del model.
        /// </summary>
        /// <typeparam name="T">Tipus del camp o la propietat que es vol modificar.</typeparam>
        /// <param name="field">Referència al camp o a la propietat que es vol modificar.</param>
        /// <param name="value">Nou valor del camp o la propietat a modificar.</param>
        /// <param name="expression">
        /// Expressió lambda que retorna el camp o la propietat que es vol modificar.
        /// </param>
        /// <returns>
        /// Vertader si s'ha completat amb èxit el canvi de valor del camp o de la propietat.
        /// Fals si el canvi de valor no ha passat el procés d'acceptació del nou valor.
        /// </returns>
        /// <remarks>En cas d'errors no control·lats es propaguen les excepcions.</remarks>
        protected virtual bool SetField<T>(ref T field, T value, Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de les implementacions
            // de la classe i en retorna el resultat.
            return SetField<T>(ref field, value, body.Member.Name);
        }

        #endregion

        #endregion

    }

}
