﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Linq.Expressions;
using System.Runtime.CompilerServices;

namespace fgxavi.Mvvm
{

    #region delegate ModelBindingEventHandler

    /// <summary>
    /// Delegat que encapsula els mètodes dels events <see cref="ModelCollectionBase{T}.Inserting"/>,
    /// <see cref="ModelCollectionBase{T}.Updating"/> i <see cref="ModelCollectionBase{T}.Deleting"/>
    /// de la classe <see cref="ModelCollectionBase{T}"/>.
    /// </summary>
    /// <param name="sender">
    /// Model de dades <see cref="ModelBase"/> que produeix l'event.
    /// </param>
    public delegate void ModelBindingEventHandler(ModelBase sender);

    #endregion

    /// <summary>
    /// Generalització de la classe <see cref="ObservableCollection"/>
    /// que implementa la interfície <see cref="IEditableObject"/> per fer editable
    /// la col·lecció d'objectes i alliberar la capa de dades de la capa de negoci.
    /// </summary>
    public class ModelCollectionBase<T> : ObservableCollection<T>,
        INotifyPropertyChanged, IChangeTracking, IRevertibleChangeTracking,
        ISelectable, ICollectionSelectable<T>
        where T : ModelBase
    {

        #region Member _lock

        /// <summary>
        /// Objecte privat de control que s'utilitza per bloquejar
        /// el model de dades de possibles processos concurrents.
        /// </summary>
        private readonly object _lock = new object();

        #endregion

        #region Property HasDetached

        /// <summary>
        /// Informa si existeix algun model de dades nou sense modificar per no crear-ne més.
        /// És vertader si hi ha algun model de dades nou sense modificar i fals si no és així.
        /// </summary>
        public bool HasDetached
        {
            get { return Items.Any(i => i.State == DataRowState.Detached); }
            protected set
            {
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Property IsEmpty

        /// <summary>
        /// Informa si la col·lecció no conté cap model de dades.
        /// Vertader si no hi ha cap model de dades a la col·lecció i fals si no és així.
        /// </summary>
        public bool IsEmpty
        {
            get { return Items.Count == 0; }
            protected set
            {
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Method AddNew

        /// <summary>
        /// Afegeix un model de dades nou a la col·lecció.
        /// </summary>
        public void AddNew()
        {
            // S'instancia un model de dades nou amb el seu constructor per defecte.
            T item = Activator.CreateInstance(typeof(T)) as T;
            // S'inserta l'element a la col·lecció.
            Add(item);
            // Selecciona el model de dades.
            item.Selected = true;
            // Comença l'edició del model de dades.
            item.BeginEdit();
        }

        #endregion

        #region RemoveSelected

        /// <summary>
        /// Esborra els models de dades seleccionats de la col·lecció.
        /// </summary>
        /// <remarks>
        /// Tan sols deixa els models de dades en estat Deleted.
        /// Els models de dades en estat Detached o Added són eliminats de la col·lecció.
        /// </remarks>
        public void RemoveSelected()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // S'obtenen els elements seleccionats.
                var selected = Items.Where(i => i.Selected);
                // Si no hi ha elements seleccionats surt.
                if (selected.Count() == 0)
                    return;
                // S'apliquen els canvis a tots els models seleccionats.
                foreach (T item in selected.ToList())
                {
                    switch (item.State)
                    {
                        case DataRowState.Detached:
                        case DataRowState.Added:
                            // S'elimina de la col·lecció els models de dades afegits.
                            Remove(item);
                            break;
                        case DataRowState.Deleted:
                        case DataRowState.Modified:
                        case DataRowState.Unchanged:
                        default:
                            // Canvia l'estat dels models de dades a esborrat
                            // per notificar-ho a la capa de negoci quan s'apliquin els canvis.
                            item.State = DataRowState.Deleted;
                            break;
                    }
                }
                // S'actualitzen les propietats afectades.
                HasDetached = false;
                IsChanged = false;
            }
        }

        #endregion

        #region RemoveAll

        /// <summary>
        /// Esborra els models de dades seleccionats de la col·lecció.
        /// </summary>
        /// <remarks>
        /// Tan sols deixa els models de dades en estat Deleted.
        /// Els models de dades en estat Detached o Added són eliminats de la col·lecció.
        /// </remarks>
        public void RemoveAll()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Si no hi ha elements surt.
                if (Items.Count() == 0)
                    return;
                // S'apliquen els canvis a tots els models seleccionats.
                foreach (T item in Items.ToList())
                {
                    switch (item.State)
                    {
                        case DataRowState.Detached:
                        case DataRowState.Added:
                            // S'elimina de la col·lecció els models de dades afegits.
                            Remove(item);
                            break;
                        case DataRowState.Deleted:
                        case DataRowState.Modified:
                        case DataRowState.Unchanged:
                        default:
                            // Canvia l'estat dels models de dades a esborrat
                            // per notificar-ho a la capa de negoci quan s'apliquin els canvis.
                            item.State = DataRowState.Deleted;
                            break;
                    }
                }
                // S'actualitzen les propietats afectades.
                HasDetached = false;
                IsChanged = false;
            }
        }

        #endregion

        #region Constructors

        public ModelCollectionBase() : base() { }

        public ModelCollectionBase(IEnumerable<T> collection)
            : base(collection)
        {
            // S'inicialitza correctament cada element a la col·lecció.
            foreach (T item in collection)
            {
                // Assigna els controladors d'events al model de dades.
                SetControllers(item);
                // S'afegeix el model de dades sense modificacions.
                item.State = DataRowState.Unchanged;
            }
        }

        public ModelCollectionBase(IList<T> list)
            : base(list)
        {
            // S'inicialitza correctament cada element a la col·lecció.
            foreach (T item in list)
            {
                // Assigna els controladors d'events al model de dades.
                SetControllers(item);
                // S'afegeix el model de dades sense modificacions.
                item.State = DataRowState.Unchanged;
            }
        }

        #endregion

        #region Events

        /// <summary>
        /// Event que es produeix quan es comença a editar l'objecte.
        /// </summary>
        public event ModelEditEventHandler ModelBeginEdit;

        /// <summary>
        /// Event que es produeix quan es s'acaba d'editar l'objecte.
        /// </summary>
        public event ModelEditEventHandler ModelEndEdit;

        /// <summary>
        /// Event que es produeix quan s'inserta un element a la col·lecció.
        /// </summary>
        public event ModelBindingEventHandler Inserting;

        /// <summary>
        /// Event que es produeix quan es modifica un element de la col·lecció.
        /// </summary>
        public event ModelBindingEventHandler Updating;

        /// <summary>
        /// Event que es produeix quan s'esborra un element de la col·lecció.
        /// </summary>
        public event ModelBindingEventHandler Deleting;

        #endregion

        #region Override Method InsertItem

        /// <summary>
        /// Sobrecarrega el mètode <see cref="ObservableCollection{T}.InsertItem"/>
        /// per assignar els controladors dels event dels models de dades
        /// afegits a la col·lecció.
        /// </summary>
        /// <param name="index">Posició del element a la col·lecció.</param>
        /// <param name="item">Element afegit a la col·lecció.</param>
        protected override void InsertItem(int index, T item)
        {
            // S'inserta l'element a la posició especificada.
            base.InsertItem(index, item);
            // Assigna els controladors d'events al model de dades.
            SetControllers(item);
            // S'afegeix el model de dades com a nou.
            item.State = DataRowState.Detached;
            // S'actualitza la propietat HasDetached.
            HasDetached = true;
        }

        #endregion

        #region Method SetControllers

        /// <summary>
        /// Assigna els controladors d'events a un model de dades.
        /// </summary>
        /// <param name="item">Model de dades.</param>
        protected void SetControllers(T item)
        {
            // S'assigna un controlador per l'event ModelBeginEdit del element.
            item.ModelBeginEdit += OnModelBeginEdit;
            // S'assigna un controlador per l'event ModelEndEdit del element.
            item.ModelEndEdit += OnModelEndEdit;
            // S'assigna un controlador per l'event SendAcceptChanges del element.
            item.SendAcceptChanges += (model) => OnSendAcceptChanges(model as T);
            // S'assigna un controlador per l'event SendAcceptChanges del element.
            item.RecieveAcceptChanges += OnRecieveAcceptChanges;
            // S'assigna un controlador per l'event SendRejectChanges del element.
            item.SendRejectChanges += (model) => OnSendRejectChanges(model as T);
            // S'assigna un controlador per l'event SelectionChanged del element.
            item.SelectionChanged += OnSelectionChanged;
            // S'emmagatzema una còpia del estat actual del model de dades.
            item.Backup();
        }

        #endregion

        #region Controllers

        #region Controller OnModelBeginEdit

        /// <summary>
        /// Controlador que desencadena l'event <see cref="ModelBeginEdit"/>
        /// de la col·lecció d'elements editables.
        /// </summary>
        /// <param name="sender">
        /// Model de dades <see cref="ModelBase"/> que produeix l'event.
        /// </param>
        protected void OnModelBeginEdit(ModelBase sender)
        {
            if (ModelBeginEdit != null)
                ModelBeginEdit(sender);
        }

        #endregion

        #region Controller OnModelEndEdit

        /// <summary>
        /// Controlador que desencadena l'event <see cref="ModelEndEdit"/>
        /// de la col·lecció d'elements editables.
        /// </summary>
        /// <param name="sender">
        /// Model de dades <see cref="ModelBase"/> que produeix l'event.
        /// </param>
        protected void OnModelEndEdit(ModelBase sender)
        {
            if (ModelEndEdit != null)
            {
                ModelEndEdit(sender);
                // Desencadena l'event ItemUpdating per actualitzar el model a la base de dades.
                Updating(sender as ModelBase);
            }
        }

        #endregion

        #region Override Controller OnCollectionChanged

        /// <summary>
        /// Sobrecarrega el Controlador de canvis de la col·lecció.
        /// </summary>
        /// <param name="e">Informació dels canvis de la col·lecció.</param>
        protected override void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            // Crida al Controlador base que desencadena l'event CollectionChanged.
            base.OnCollectionChanged(e);
            // Identifica l'acció realitzada per desencadenar l'event corresponent.
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    // Desencadena un event ItemInserting per cada element insertat.
                    if (Inserting != null)
                        foreach (ModelBase item in e.NewItems)
                            Inserting(item);
                    break;
                case NotifyCollectionChangedAction.Move:
                    break;
                case NotifyCollectionChangedAction.Remove:
                    // Desencadena un event ItemDeleting per cada element eliminat.
                    if (Deleting != null)
                        foreach (ModelBase item in e.OldItems)
                            Deleting(item);
                    break;
                case NotifyCollectionChangedAction.Replace:
                    break;
                case NotifyCollectionChangedAction.Reset:
                    break;
                default:
                    break;
            }
        }

        #endregion

        #endregion

        #region INotifyPropertyChanged implementation

        #region Event INotifyPropertyChanged.PropertyChanged

        /// <summary>
        /// Event que es produeix quan es modifica una propietat.
        /// </summary>
        public new event PropertyChangedEventHandler PropertyChanged;

        #endregion

        #region Controller RaisePropertyChanged([CallerMemberName]string propertyName = null)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanged"/>
        /// un cop s'ha modificat un camp o una propietat.
        /// </summary>
        /// <param name="propertyName">
        /// Nom de la propietat que s'ha modificat.
        /// </param>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        protected virtual void RaisePropertyChanged([CallerMemberName]string propertyName = null)
        {
            // Si hi ha un Controlador enllaçat a l'event PropertyChanged l'executa.
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

        #region Controller RaisePropertyChanged<T>(Expression<Func<T>> expression)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanged"/>
        /// un cop s'ha modificat un camp o una propietat.
        /// </summary>
        /// <typeparam name="T">
        /// Tipus retornat per l'expressió lambda del paràmetre <paramref name="expression"/>.
        /// </typeparam>
        /// <param name="expression">
        /// Expressió lambda que ha de retornar el camp o la propietat que s'ha modificat.
        /// </param>
        /// <remarks>
        /// Si l'expressió lambda no retorna un camp o una propietat es produeixen excepcions.
        /// </remarks>
        protected virtual void RaisePropertyChanged<TT>(Expression<Func<TT>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de la implementació INotifyPropertyChanged.
            RaisePropertyChanged(body.Member.Name);
        }

        #endregion

        #endregion

        #region IChangeTracking implementation

        #region Events SendAcceptChanges

        /// <summary>
        /// Event que es produeix quan s'accepten els canvis
        /// d'un model de dades nou.
        /// </summary>
        public event SendChangesEventHandler<T> SendAcceptChanges;

        #endregion

        #region Events RecieveAcceptChanges

        /// <summary>
        /// Event que es produeix quan s'accepten els canvis
        /// d'un model de dades nou.
        /// </summary>
        public event RecieveChangesEventHandler<ModelBase> RecieveAcceptChanges;

        #endregion

        #region Controllers AcceptChanges

        #region Controller OnSendAcceptChanges

        /// <summary>
        /// Controlador que desencadena l'event <see cref="AcceptInsertChanges"/>
        /// de la col·lecció de models de dades.
        /// </summary>
        /// <param name="sender">Model de dades que produeix l'event.</param>
        protected void OnSendAcceptChanges(T sender)
        {
            if (SendAcceptChanges != null)
                SendAcceptChanges(sender);
        }

        #endregion

        #region Controller OnRecieveAcceptChanges

        /// <summary>
        /// Controlador que desencadena l'event <see cref="AcceptInsertChanges"/>
        /// de la col·lecció de models de dades.
        /// </summary>
        /// <param name="sender">Model de dades que produeix l'event.</param>
        protected void OnRecieveAcceptChanges(ModelBase sender, ChangeResultEventArgs<ModelBase> args)
        {
            if (RecieveAcceptChanges != null)
                RecieveAcceptChanges(sender, args);
            if (!args.Result.HasError
                && args.Result.EntityResult != null
                && args.Result.EntityResult.Entity != null
                && sender.State == DataRowState.Deleted)
            {
                if (SelectedItem == sender)
                    SelectedItem = null;
                Remove(sender as T);
            }
        }

        #endregion

        #endregion

        #region Property IChangeTracking.IsChanged

        /// <summary>
        /// Estat de canvi dels models de dades.
        /// És vertader si els models de dades s'han modificat i fals si no és així.
        /// </summary>
        public bool IsChanged
        {
            get { return Items.Any(i => i.IsChanged); }
            protected set
            {
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Property IsSelectedChanged

        /// <summary>
        /// Estat de canvi de la selecció del model de dades.
        /// És vertader si els models de dades seleccionats s'han modificat i fals si no és així.
        /// </summary>
        public bool IsSelectedChanged
        {
            get { return Items.Any(i => i.Selected && i.IsChanged); }
            protected set
            {
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Method IChangeTracking.AcceptChanges

        /// <summary>
        /// Desencadena l'event <see cref="SendAcceptChanges"/> de cada model modificat per sol·licitar
        /// a la capa de la lògica de negoci que apliqui els canvis dels models de dades.
        /// </summary>
        public void AcceptChanges()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // S'apliquen els canvis a tots els models que s'hagin modificat.
                foreach (T item in Items.Where(i => i.IsChanged).ToList())
                    item.AcceptChanges();
                IsChanged = false;
            }
        }

        #endregion

        #region Method AcceptSelectedChanges

        /// <summary>
        /// Desencadena l'event <see cref="SendAcceptChanges"/> de cada model modificat seleccionat
        /// per sol·licitar a la capa de la lògica de negoci que apliqui els canvis dels models de dades.
        /// </summary>
        public void AcceptSelectedChanges()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // S'apliquen els canvis a tots els models seleccionats que s'hagin modificat.
                foreach (T item in Items.Where(i => i.Selected && i.IsChanged).ToList())
                    item.AcceptChanges();
                IsChanged = false;
            }
        }

        #endregion

        #endregion

        #region IRevertibleChangeTracking implementation

        #region Controller OnRejectModelChanges

        /// <summary>
        /// Controlador que captura l'event <see cref="ModelBase.RejectChanges"/>
        /// de la col·lecció de models de dades per manipular el model de dades a la col·lecció.
        /// </summary>
        /// <param name="sender">Model de dades que produeix l'event.</param>
        protected void OnSendRejectChanges(T item)
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Es desencadena l'event corresponent.
                switch (item.State)
                {
                    case DataRowState.Detached:
                    case DataRowState.Added:
                        // Esborra el model de dades de la col·lecció.
                        Remove(item);
                        break;
                    case DataRowState.Modified:
                    case DataRowState.Deleted:
                    case DataRowState.Unchanged:
                    default:
                        break;
                }
            }
        }

        #endregion

        #region Method IRevertibleChangeTracking.RejectChanges

        /// <summary>
        /// Retorna els models de dades al seu estat original
        /// abans de totes les modificacions.
        /// </summary>
        public void RejectChanges()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Es restauren els canvis de tots els models que s'hagin modificat.
                foreach (T item in Items.Where(i => i.IsChanged).ToList())
                    item.RejectChanges();
                IsChanged = false;
            }
        }

        #endregion

        #region Method RejectSelectedChanges

        /// <summary>
        /// Retorna els models de dades seleccionats al seu estat original
        /// abans de totes les modificacions.
        /// </summary>
        public void RejectSelectedChanges()
        {
            // Es protegeix el model de dades amb la sentència lock i el camp de control _lock
            // per a que cap procés el modifiqui mentre aquest bloc de codi s'està executant.
            lock (_lock)
            {
                // Es restauren els canvis de tots els models seleccionats que s'hagin modificat.
                foreach (T item in Items.Where(i => i.Selected && i.IsChanged).ToList())
                    item.RejectChanges();
                IsChanged = false;
            }
        }

        #endregion

        #endregion

        #region ISelectable implementation

        #region Event ISelectable.SelectionChanged

        /// <summary>
        /// Event que es produeix quan es modifica la selecció del objecte.
        /// </summary>
        public event SelectionChangedHandler SelectionChanged;

        #endregion

        #region Controller OnSelectionChanged

        /// <summary>
        /// Controlador que desencadena l'event <see cref="SelectionChanged"/>.
        /// </summary>
        /// <param name="sender">
        /// Objecte seleccionable <see cref="ISelectable"/> que produeix l'event.
        /// </param>
        /// <param name="selection">
        /// Informa al controlador si s'ha seleccionat l'objecte o si s'ha deseleccionat.
        /// </param>
        protected void OnSelectionChanged(ISelectable sender, bool selection)
        {
            if (SelectionChanged != null)
                SelectionChanged(sender, selection);
            // Si no s'està editant un element
            // es pren com a principal l'element que es selecciona.
            if (!Items.Any(i => i.IsEditing))
                SelectedItem = sender as T;
            // Actualitza les propietats relacionades.
            RaisePropertyChanged(() => Selected);
            RaisePropertyChanged(() => HasSelection);
            RaisePropertyChanged(() => FullSelection);
        }

        #endregion

        #region Property ISelectable.Selected

        /// <summary>
        /// Membre privat de la propietat <see cref="Selected" />.
        /// </summary>
        private bool _Selected = false;

        /// <summary>
        /// Estat de selecció del objecte.
        /// És vertader si l'objecte està seleccionat i fals si no és així.
        /// Estableix i obté la propietat <see cref="Selected" />.
        /// </summary>
        public bool Selected
        {
            get { return _Selected; }
            set
            {
                if (_Selected != value)
                {
                    // S'assigna el nou valor.
                    _Selected = value;
                    // Crida el controlador de la interfície INotifyPropertyChanged.
                    RaisePropertyChanged();
                    // Crida el controlador de la interfície ISelectable.
                    OnSelectionChanged(this, _Selected);
                }
            }
        }

        #endregion

        #region Property SelectedItem

        /// <summary>
        /// Membre privat de la propietat <see cref="SelectedItem" />.
        /// </summary>
        private T _SelectedItem = null;

        /// <summary>
        /// Element principal seleccionat.
        /// Estableix i obté la propietat <see cref="SelectedItem" />.
        /// </summary>
        public T SelectedItem
        {
            get { return _SelectedItem; }
            set
            {
                if (_SelectedItem != value)
                {
                    // S'assigna el nou valor.
                    _SelectedItem = value;
                    // Crida el controlador de la interfície INotifyPropertyChanged.
                    RaisePropertyChanged();
                }
            }
        }

        #endregion

        #endregion

        #region ICollectionSelectable<T> implementation

        #region Property ICollectionSelectable<T>.SelectedItems

        /// <summary>
        /// Obté una col·lecció que conté tots els elements seleccionats.
        /// </summary>
        public Collection<T> SelectedItems
        {
            get
            {
                return new Collection<T>(Items.Where(i => i.Selected).ToList());
            }
        }

        #endregion

        #region Property ICollectionSelectable<T>.HasSelection

        /// <summary>
        /// Informa si hi ha algun element seleccionat.
        /// És vertader si hi ha algun element seleccionat i fals si no és així.
        /// </summary>
        public bool HasSelection
        {
            get { return Items.Any(i => i.Selected); }
            protected set
            {
                // Crida el Controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Property ICollectionSelectable<T>.FullSelection

        /// <summary>
        /// Informa si tots els elements estan seleccionats.
        /// Vertader si tots els elements estan seleccionats i fals si no és així.
        /// </summary>
        public bool FullSelection
        {
            get { return Items.All(i => i.Selected); }
            set
            {
                // S'obté el valor de selecció que compartiran tots els elements.
                // Si no estan tots seleccionats es seleccionen,
                // si estan tots seleccionats es deseleccionen.
                bool selected = !Items.All(i => i.Selected);
                // S'assigna el mateix estat de selecció a tots els elements.
                foreach (T item in Items)
                    item.Selected = selected;
                // Crida el controlador de la interfície INotifyPropertyChanged.
                RaisePropertyChanged();
            }
        }

        #endregion

        #endregion

    }

}
