﻿using System;
using System.Linq;
using System.Reflection;

namespace fgxavi.Mvvm
{
    public class ReflectionHelper
    {

        /// <summary>
        /// Retorna un tipus <see cref="Type"/> de l'herència d'un tipus <see cref="Type"/>.
        /// </summary>
        /// <param name="source">Tipus <see cref="Type"/> origen.</param>
        /// <param name="predicate">Condició de cerca del tipus <see cref="Type"/>.</param>
        /// <returns>Tipus <see cref="Type"/> trobat.</returns>
        /// <remarks>Si no troba el tipus <see cref="Type"/> retorna null.</remarks>
        public static Type FindParentType(Type source, Func<Type, bool> predicate)
        {
            // Si els paràmetres són nuls retorna nul.
            if (source == null || predicate == null)
                return null;
            // Recorre l'herència de tipus del tipus origen
            // cercant amb la condició de cerca facilitada.
            Type current = source;
            while (!typeof(object).Equals(current))
            {
                // Cerca positiva.
                if (predicate(current))
                    return current;
                // Analitza el tipus de la classe pare.
                current = current.BaseType;
            }
            // Cerca negativa.
            return null;
        }

        /// <summary>
        /// Retorna si existeix un tipus <see cref="Type"/> a l'herència d'un tipus <see cref="Type"/>.
        /// </summary>
        /// <param name="source">Tipus <see cref="Type"/> origen.</param>
        /// <param name="predicate">Condició de cerca del tipus <see cref="Type"/>.</param>
        /// <returns>Vertader si existeix el tipus <see cref="Type"/>,
        /// fals si no existeix.</returns>
        public static bool ContainsParentType(Type source, Func<Type, bool> predicate)
        {
            // Retorna si existeix cerca el tipus a l'herència.
            return FindParentType(source, predicate) != null;
        }

        /// <summary>
        /// Retorna les propietats <see cref="PropertyInfo"/> d'un tipus <see cref="Type"/> que compleixen una condició.
        /// </summary>
        /// <param name="source">Tipus <see cref="Type"/> origen.</param>
        /// <param name="predicate">Condició de cerca de la propietat <see cref="PropertyInfo"/>.</param>
        /// <returns>Propietats <see cref="PropertyInfo"/> trobades.</returns>
        /// <remarks>Si no troba cap propietat <see cref="PropertyInfo"/> retorna un array buit.</remarks>
        public static PropertyInfo[] FindProperties(Type source, Func<PropertyInfo, bool> predicate)
        {
            // Si els paràmetres són nuls retorna nul.
            if (source == null || predicate == null)
                return new PropertyInfo[] { };
            // Recorre l'herència de tipus del tipus origen
            // cercant amb la condició de cerca facilitada.
            return source.GetProperties().Where(p => predicate(p)).ToArray();
        }

        /// <summary>
        /// Retorna la primera propietat <see cref="PropertyInfo"/> d'un tipus <see cref="Type"/> que compleix una condició.
        /// </summary>
        /// <param name="source">Tipus <see cref="Type"/> origen.</param>
        /// <param name="predicate">Condició de cerca de la propietat <see cref="PropertyInfo"/>.</param>
        /// <returns>Propietat <see cref="PropertyInfo"/> trobada.</returns>
        /// <remarks>Si no troba cap propietat <see cref="PropertyInfo"/> retorna nul.</remarks>
        public static PropertyInfo FindProperty(Type source, Func<PropertyInfo, bool> predicate)
        {
            // Obté un array amb les propietats trobades.
            PropertyInfo[] props = FindProperties(source, predicate);
            // Si s'obtenen propietats retorna només la primera, sinò retorna nul.
            return props.Length > 0 ? props[0] : null;
        }

    }

}
