﻿using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Messaging;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics.CodeAnalysis;
using System.Linq.Expressions;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Windows;

namespace fgxavi.Mvvm
{
    /// <summary>
    /// Vista abstracta d'un model de dades
    /// del patró de disseny MVVM.
    /// </summary>
    public abstract class ViewModelBase : ModelBase, ICleanup
    {

        #region Properties

        #region Property MessengerInstance

        /// <summary>
        /// Membre de la propietat <see cref="MessengerInstance"/>.
        /// </summary>
        private IMessenger _messengerInstance;

        /// <summary>
        /// Obté o estableix una instància d'un
        /// <see cref="GalaSoft.MvvmLight.Messaging.IMessenger"/>
        /// utilitzat per transmetre missatges a altres objectes.
        /// </summary>
        /// <remarks>
        /// Si és null, aquesta classe tractarà de transmetre
        /// mitjançant la instància per defecte del Messenger.
        /// </remarks>
        protected IMessenger MessengerInstance
        {
            get { return _messengerInstance ?? Messenger.Default; }
            set { _messengerInstance = value; }
        }

        #endregion

        #region Property Token

        /// <summary>
        /// Propietat de control que identifica l'objecte
        /// per a rebre missatges de <see cref="IMessenger"/>.
        /// </summary>
        public object Token { get; set; }

        #endregion

        #region Property IsInDesignModeStatic

        /// <summary>
        /// Membre estàtic de la propietat <see cref="IsInDesignMode"/>
        /// </summary>
        private static bool? _isInDesignMode;

        /// <summary>
        /// Indica si el contexte està en mode de disseny 
        /// amb Blend o Visual Studio o en mode execució.
        /// </summary>
        [SuppressMessage(
            "Microsoft.Security",
            "CA2122:DoNotIndirectlyExposeMethodsWithLinkDemands",
            Justification = "The security risk here is neglectible.")]
        public static bool IsInDesignModeStatic
        {
            get
            {
                if (!_isInDesignMode.HasValue)
                {
                    var prop = DesignerProperties.IsInDesignModeProperty;
                    _isInDesignMode = (bool)DependencyPropertyDescriptor
                        .FromProperty(prop, typeof(FrameworkElement))
                        .Metadata.DefaultValue;
                }
                return _isInDesignMode.Value;
            }
        }

        #endregion

        #region Property IsInDesignMode

        /// <summary>
        /// Indica si el contexte està en mode de disseny 
        /// amb Blend o Visual Studio o en mode execució.
        /// </summary>
        [SuppressMessage(
            "Microsoft.Performance",
            "CA1822:MarkMembersAsStatic",
            Justification = "Non static member needed for data binding")]
        public bool IsInDesignMode
        {
            get
            {
                return IsInDesignModeStatic;
            }
        }

        #endregion

        #endregion

        #region Commands

        #region RelayCommand RefreshCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RefreshCommand" />.
        /// </summary>
        private RelayCommand _RefreshCommand;

        /// <summary>
        /// Obté la commanda <see cref="RefreshCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand RefreshCommand
        {
            get
            {
                return _RefreshCommand
                    ?? (_RefreshCommand = new RelayCommand(
                        ExecuteRefreshCommand,
                        CanExecuteRefreshCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RefreshCommand" />.
        /// </summary>
        protected virtual void ExecuteRefreshCommand()
        {
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RefreshCommand" />.
        /// </summary>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteRefreshCommand()
        {
            return false;
        }

        #endregion

        #region RelayCommand<ModelBase> EditCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="EditCommand" />.
        /// </summary>
        private RelayCommand<ModelBase> _EditCommand;

        /// <summary>
        /// Obté la commanda <see cref="EditCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<ModelBase> EditCommand
        {
            get
            {
                return _EditCommand
                    ?? (_EditCommand = new RelayCommand<ModelBase>(
                        ExecuteEditCommand,
                        CanExecuteEditCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="EditCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        protected virtual void ExecuteEditCommand(ModelBase model)
        {
            model.BeginEdit();
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="EditCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteEditCommand(ModelBase model)
        {
            return model != null && !model.IsEditing;
        }

        #endregion

        #region RelayCommand<ModelBase> CancelCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="CancelCommand" />.
        /// </summary>
        private RelayCommand<ModelBase> _CancelCommand;

        /// <summary>
        /// Obté la commanda <see cref="CancelCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<ModelBase> CancelCommand
        {
            get
            {
                return _CancelCommand
                    ?? (_CancelCommand = new RelayCommand<ModelBase>(
                        ExecuteCancelCommand,
                        CanExecuteCancelCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="CancelCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        protected virtual void ExecuteCancelCommand(ModelBase model)
        {
            model.CancelEdit();
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="CancelCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteCancelCommand(ModelBase model)
        {
            return model != null && model.IsEditing;
        }

        #endregion

        #region RelayCommand<ModelBase> SaveCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="SaveCommand" />.
        /// </summary>
        private RelayCommand<ModelBase> _SaveCommand;

        /// <summary>
        /// Obté la commanda <see cref="SaveCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<ModelBase> SaveCommand
        {
            get
            {
                return _SaveCommand
                    ?? (_SaveCommand = new RelayCommand<ModelBase>(
                        ExecuteSaveCommand,
                        CanExecuteSaveCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="SaveCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        protected virtual void ExecuteSaveCommand(ModelBase model)
        {
            model.EndEdit();
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="SaveCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteSaveCommand(ModelBase model)
        {
            return model != null && model.IsEditing;
        }

        #endregion

        #region RelayCommand<ModelBase> AcceptChangesCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="AcceptChangesCommand" />.
        /// </summary>
        private RelayCommand<ModelBase> _AcceptChangesCommand;

        /// <summary>
        /// Obté la commanda <see cref="AcceptChangesCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<ModelBase> AcceptChangesCommand
        {
            get
            {
                return _AcceptChangesCommand
                    ?? (_AcceptChangesCommand = new RelayCommand<ModelBase>(
                        ExecuteAcceptChangesCommand,
                        CanExecuteAcceptChangesCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="AcceptChangesCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        protected virtual void ExecuteAcceptChangesCommand(ModelBase model)
        {
            model.AcceptChanges();
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="AcceptChangesCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteAcceptChangesCommand(ModelBase model)
        {
            return model != null && !model.IsEditing && model.IsChanged;
        }

        #endregion

        #region RelayCommand<ModelBase> RejectChangesCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RejectChangesCommand" />.
        /// </summary>
        private RelayCommand<ModelBase> _RejectChangesCommand;

        /// <summary>
        /// Obté la commanda <see cref="RejectChangesCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<ModelBase> RejectChangesCommand
        {
            get
            {
                return _RejectChangesCommand
                    ?? (_RejectChangesCommand = new RelayCommand<ModelBase>(
                        ExecuteRejectChangesCommand,
                        CanExecuteRejectChangesCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RejectChangesCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        protected virtual void ExecuteRejectChangesCommand(ModelBase model)
        {
            model.RejectChanges();
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RejectChangesCommand" />.
        /// </summary>
        /// <param name="model">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteRejectChangesCommand(ModelBase model)
        {
            return model != null && !model.IsEditing && model.IsChanged;
        }

        #endregion

        #region RelayCommand<object> AddCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="AddCommand" />.
        /// </summary>
        private RelayCommand<object> _AddCommand;

        /// <summary>
        /// Obté la commanda <see cref="AddCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> AddCommand
        {
            get
            {
                return _AddCommand
                    ?? (_AddCommand = new RelayCommand<object>(
                        (collection) => ExecuteModelCollectionCommand(collection, "AddNew"),
                        (collection) => !CanExecuteModelCollectionCommand(collection, "HasDetached")));
            }
        }

        #endregion

        #region RelayCommand<object> AcceptSelectedChangesCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="AcceptSelectedChangesCommand" />.
        /// </summary>
        private RelayCommand<object> _AcceptSelectedChangesCommand;

        /// <summary>
        /// Obté la commanda <see cref="AcceptSelectedChangesCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> AcceptSelectedChangesCommand
        {
            get
            {
                return _AcceptSelectedChangesCommand
                    ?? (_AcceptSelectedChangesCommand = new RelayCommand<object>(
                        ExecuteAcceptSelectedChangesCommand,
                        CanExecuteAcceptSelectedChangesCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="AcceptSelectedChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        protected virtual void ExecuteAcceptSelectedChangesCommand(object collection)
        {
            ExecuteModelCollectionCommand(collection, "AcceptSelectedChanges");
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="AcceptSelectedChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteAcceptSelectedChangesCommand(object collection)
        {
            return CanExecuteModelCollectionCommand(collection, "IsSelectedChanged")
                            && !CanExecuteModelCollectionCommand(collection, "FullSelection");
        }

        #endregion

        #region RelayCommand<object> RejectSelectedChangesCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RejectSelectedChangesCommand" />.
        /// </summary>
        private RelayCommand<object> _RejectSelectedChangesCommand;

        /// <summary>
        /// Obté la commanda <see cref="RejectSelectedChangesCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> RejectSelectedChangesCommand
        {
            get
            {
                return _RejectSelectedChangesCommand
                    ?? (_RejectSelectedChangesCommand = new RelayCommand<object>(
                        ExecuteRejectSelectedChangesCommand,
                        CanExecuteRejectSelectedChangesCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RejectSelectedChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        protected virtual void ExecuteRejectSelectedChangesCommand(object collection)
        {
            ExecuteModelCollectionCommand(collection, "RejectSelectedChanges");
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RejectSelectedChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteRejectSelectedChangesCommand(object collection)
        {
            return CanExecuteModelCollectionCommand(collection, "IsSelectedChanged")
                            && !CanExecuteModelCollectionCommand(collection, "FullSelection");
        }

        #endregion

        #region RelayCommand<object> AcceptAllChangesCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="AcceptAllChangesCommand" />.
        /// </summary>
        private RelayCommand<object> _AcceptAllChangesCommand;

        /// <summary>
        /// Obté la commanda <see cref="AcceptAllChangesCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> AcceptAllChangesCommand
        {
            get
            {
                return _AcceptAllChangesCommand
                    ?? (_AcceptAllChangesCommand = new RelayCommand<object>(
                        ExecuteAcceptAllChangesCommand,
                        CanExecuteAcceptAllChangesCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="AcceptAllChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        protected virtual void ExecuteAcceptAllChangesCommand(object collection)
        {
            ExecuteModelCollectionCommand(collection, "AcceptChanges");
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="AcceptAllChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteAcceptAllChangesCommand(object collection)
        {
            return CanExecuteModelCollectionCommand(collection, "IsChanged");
        }

        #endregion

        #region RelayCommand<object> RejectAllChangesCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RejectAllChangesCommand" />.
        /// </summary>
        private RelayCommand<object> _RejectAllChangesCommand;

        /// <summary>
        /// Obté la commanda <see cref="RejectAllChangesCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> RejectAllChangesCommand
        {
            get
            {
                return _RejectAllChangesCommand
                    ?? (_RejectAllChangesCommand = new RelayCommand<object>(
                        ExecuteRejectAllChangesCommand,
                        CanExecuteRejectAllChangesCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RejectAllChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        protected virtual void ExecuteRejectAllChangesCommand(object collection)
        {
            ExecuteModelCollectionCommand(collection, "RejectChanges");
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RejectAllChangesCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteRejectAllChangesCommand(object collection)
        {
            return CanExecuteModelCollectionCommand(collection, "IsChanged");
        }

        #endregion

        #region RelayCommand<object> RemoveSelectedCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RemoveSelectedCommand" />.
        /// </summary>
        private RelayCommand<object> _RemoveSelectedCommand;

        /// <summary>
        /// Obté la commanda <see cref="RemoveSelectedCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> RemoveSelectedCommand
        {
            get
            {
                return _RemoveSelectedCommand
                    ?? (_RemoveSelectedCommand = new RelayCommand<object>(
                        ExecuteRemoveSelectedCommand,
                        CanExecuteRemoveSelectedCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RemoveSelectedCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        protected virtual void ExecuteRemoveSelectedCommand(object collection)
        {
            ExecuteModelCollectionCommand(collection, "RemoveSelected");
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RemoveSelectedCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteRemoveSelectedCommand(object collection)
        {
            return CanExecuteModelCollectionCommand(collection, "HasSelection")
                && !CanExecuteModelCollectionCommand(collection, "FullSelection");
        }

        #endregion

        #region RelayCommand<object> RemoveAllCommand

        /// <summary>
        /// Membre privat de la commanda <see cref="RemoveAllCommand" />.
        /// </summary>
        private RelayCommand<object> _RemoveAllCommand;

        /// <summary>
        /// Obté la commanda <see cref="RemoveAllCommand" /> amb la seva condició d'execució.
        /// </summary>
        /// <remarks>
        /// Utilitza mètodes per definir la commanda i la condició d'execució.
        /// </remarks>
        public RelayCommand<object> RemoveAllCommand
        {
            get
            {
                return _RemoveAllCommand
                    ?? (_RemoveAllCommand = new RelayCommand<object>(
                        ExecuteRemoveAllCommand,
                        CanExecuteRemoveAllCommand));
            }
        }

        /// <summary>
        /// Mètode que s'executa amb la commanda <see cref="RemoveAllCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        protected virtual void ExecuteRemoveAllCommand(object collection)
        {
            ExecuteModelCollectionCommand(collection, "RemoveAll");
        }

        /// <summary>
        /// Condició d'execució de la commanda <see cref="RemoveAllCommand" />.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        protected virtual bool CanExecuteRemoveAllCommand(object collection)
        {
            return !CanExecuteModelCollectionCommand(collection, "IsEmpty");
        }

        #endregion

        #endregion

        #region Constructors

        #region Constructor default

        /// <summary>
        /// Inicialitza una nova instància de la classe <see cref="ViewModelBase"/>.
        /// </summary>
        public ViewModelBase()
        {
            // Obté el model de dades del servei.
            //ExecuteRefreshCommand();
        }

        #endregion

        #region Constructor with IMessenger

        /// <summary>
        /// Inicialitza una nova instància de la classe <see cref="ViewModelBase"/>.
        /// </summary>
        /// <param name="messenger">
        /// Instància de la classe <see cref="GalaSoft.MvvmLight.Messaging.Messenger"/>
        /// que s'utilitza per transmetre missatges a altres objectes.
        /// </param>
        public ViewModelBase(IMessenger messenger) : base()
        {
            MessengerInstance = messenger;
        }

        #endregion

        #endregion

        #region ICleanup implementation

        /// <summary>
        /// Anul·la el registre d'aquesta instància de la classe Messenger.
        /// </summary>
        /// <remarks>
        /// Si es sobrecarrega el mètode per alliberar recursos de la subclasse
        /// s'ha de cridar a aquest mètode base per alliberar la instància de Messenger.
        /// </remarks>
        public virtual void Cleanup()
        {
            MessengerInstance.Unregister(this);
        }

        #endregion

        #region Broadcast implementation

        #region Method Broadcast<T>(T oldValue, T newValue, [CallerMemberName]string propertyName = null)

        /// <summary>
        /// Difon un PropertyChangedMessage utilitzant la instància del missatger
        /// que es va passar a aquesta classe (si està disponible)
        /// o la instància per defecte del missatger.
        /// </summary>
        /// <typeparam name="T">El tipus de la propietat que ha canviat.</typeparam>
        /// <param name="oldValue">El valor de la propietat abans del canvi.</param>
        /// <param name="newValue">El valor de la propietat després del canvi.</param>
        /// <param name="propertyName">El nom de la propietat que ha canviat.</param>
        protected virtual void Broadcast<T>(T oldValue, T newValue, [CallerMemberName]string propertyName = null)
        {
            // Crea un missatge PropertyChangedMessage amb la informació del canvi de valor de la propietat.
            var message = new PropertyChangedMessage<T>(this, oldValue, newValue, propertyName);
            // Envia el missatge a la resta d'objectes mitjançant el missatger.
            if (Token == null)
                // Si no s'ha especificat la propietat de control Token
                // s'envia el missatge a tots els objectes.
                MessengerInstance.Send(message);
            else
                // Si s'ha especificat la propietat de control Token
                // s'envia el missatge només a aquells objectes
                // que escolten a aquest objecte.
                MessengerInstance.Send(message, Token);
        }

        #endregion

        #region Controller RaisePropertyChanged<T>(T oldValue, T newValue, bool broadcast, [CallerMemberName]string propertyName = null)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanged"/>
        /// un cop s'ha modificat un camp o una propietat i envia un missatge
        /// mitjançant el missatger per notificar-ho a la resta d'objectes.
        /// </summary>
        /// <typeparam name="T">El tipus de la propietat que ha canviat.</typeparam>
        /// <param name="oldValue">El valor de la propietat abans del canvi.</param>
        /// <param name="newValue">El valor de la propietat després del canvi.</param>
        /// <param name="propertyName">El nom de la propietat que ha canviat.</param>
        /// <remarks>
        /// Si no s'especifica el camp o la propietat, l'atribut <see cref="CallerMemberNameAttibute"/>
        /// obté el nom de la propietat que crida el Controlador.
        /// </remarks>
        [SuppressMessage("Microsoft.Design", "CA1030:UseEventsWhereAppropriate",
            Justification = "This cannot be an event")]
        protected virtual void RaisePropertyChanged<T>(T oldValue, T newValue, bool broadcast, [CallerMemberName]string propertyName = null)
        {
            // Es desencadena l'event PropertyChanged amb el controlador de la classe base.
            RaisePropertyChanged(propertyName);
            // Si es sol·licita difondre un missatge es difonen els valors anterior i actual de la propietat.
            if (broadcast)
                Broadcast(oldValue, newValue, propertyName);
        }

        #endregion

        #region Controller RaisePropertyChanged<T>(T oldValue, T newValue, bool broadcast, Expression<Func<T>> expression)

        /// <summary>
        /// Controlador que desencadena l'event <see cref="PropertyChanged"/>
        /// un cop s'ha modificat un camp o una propietat i envia un missatge
        /// mitjançant el missatger per notificar-ho a la resta d'objectes.
        /// </summary>
        /// <typeparam name="T">El tipus de la propietat que ha canviat.</typeparam>
        /// <param name="oldValue">El valor de la propietat abans del canvi.</param>
        /// <param name="newValue">El valor de la propietat després del canvi.</param>
        /// <param name="expression">
        /// Expressió lambda que ha de retornar el camp o la propietat que s'ha modificat.
        /// </param>
        /// <remarks>
        /// Si l'expressió lambda no retorna un camp o una propietat es produeixen excepcions.
        /// </remarks>
        [SuppressMessage("Microsoft.Design", "CA1030:UseEventsWhereAppropriate",
            Justification = "This cannot be an event")]
        [SuppressMessage(
            "Microsoft.Design",
            "CA1006:GenericMethodsShouldProvideTypeParameter",
            Justification = "This syntax is more convenient than other alternatives.")]
        protected virtual void RaisePropertyChanged<T>(T oldValue, T newValue, bool broadcast, Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de la implementació RaisePropertyChanged amb broadcast.
            RaisePropertyChanged(oldValue, newValue, broadcast, body.Member.Name);
        }

        #endregion

        #endregion

        #region Helper methods

        #region Helper Method SetField<T>(ref T field, T newValue, [CallerMemberName]string propertyName = null)

        /// <summary>
        /// Mètode d'ajut al programador que fa ús de totes les implementacions de la classe
        /// <see cref="ModelBase"/> per afegir-les a cada propietat del model.
        /// </summary>
        /// <typeparam name="T">Tipus del camp o la propietat que es vol modificar.</typeparam>
        /// <param name="field">Referència al camp o a la propietat que es vol modificar.</param>
        /// <param name="newValue">Nou valor del camp o la propietat a modificar.</param>
        /// <param name="propertyName">Nom del camp o la propietat que es vol modificar.</param>
        /// <returns>
        /// Vertader si s'ha completat amb èxit el canvi de valor del camp o de la propietat.
        /// Fals si el canvi de valor no ha passat el procés d'acceptació del nou valor.
        /// </returns>
        /// <remarks>En cas d'errors no control·lats es propaguen les excepcions.</remarks>
        protected virtual bool SetField<T>(ref T field, T newValue, bool broadcast, [CallerMemberName]string propertyName = null)
        {
            // Si el valor del camp o propietat és el mateix no es modifica.
            if (EqualityComparer<T>.Default.Equals(field, newValue))
                // Informa que no s'ha modificat el camp o la propietat.
                return false;
            // Valida el nou valor de la propietat cercant possibles errors,
            // cridant al Controlador de la interfície INotifyDataErrorInfo
            ValidateMemberValue(newValue, propertyName);
            // Crida el Controlador de la interfície IAcceptPendingChange,
            // si s'accepta el canvi de valor es realitza.
            if (RaiseAcceptPendingChange(newValue, propertyName))
            {
                // Crida el Controlador de la interfície INotifyPropertyChanging.
                RaisePropertyChanging(propertyName);
                // S'emmagatzema el valor actual.
                var oldValue = field;
                // S'assigna el nou valor.
                field = newValue;
                // Crida el controlador RaisePropertyChanged amb difusió de missatges.
                RaisePropertyChanged(oldValue, newValue, broadcast, propertyName);
                // Informa que s'ha modificat el camp o la propietat.
                return true;
            }
            // Informa que no s'ha acceptat el canvi de valor del camp o la propietat.
            return false;
        }

        #endregion

        #region Helper Method SetField<T>(ref T field, T newValue, bool broadcast, Expression<Func<T>> expression)

        /// <summary>
        /// Mètode d'ajut al programador que fa ús de totes les implementacions de la classe
        /// <see cref="ModelBase"/> per afegir-les a cada propietat del model.
        /// </summary>
        /// <typeparam name="T">Tipus del camp o la propietat que es vol modificar.</typeparam>
        /// <param name="field">Referència al camp o a la propietat que es vol modificar.</param>
        /// <param name="newValue">Nou valor del camp o la propietat a modificar.</param>
        /// <param name="expression">
        /// Expressió lambda que retorna el camp o la propietat que es vol modificar.
        /// </param>
        /// <remarks>En cas d'errors no control·lats es propaguen les excepcions.</remarks>
        protected virtual bool SetField<T>(ref T field, T newValue, bool broadcast, Expression<Func<T>> expression)
        {
            // Si l'expressió lambda està buida genera una excepció.
            if (expression == null)
                throw new ArgumentNullException();
            // Si l'expressió lambda no conté un camp o una propietat genera una excepció.
            MemberExpression body = expression.Body as MemberExpression;
            if (body == null)
                throw new InvalidCastException();
            // Crida al Controlador principal de les implementacions
            // de la classe i en retorna el resultat.
            return SetField<T>(ref field, newValue, broadcast, body.Member.Name);
        }

        #endregion

        #region Helper Method ExecuteModelCollectionCommand

        /// <summary>
        /// Mètode que invoca un mètode d'un objecte <see cref="ModelCollectionBase" />
        /// mitjançant <see cref="Reflection"/>.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <param name="methodName">Nom del mètode a invocar.</param>
        private void ExecuteModelCollectionCommand(object collection, string methodName)
        {
            // Si el paràmetre de la comanda és nul surt.
            if (collection == null)
                return;
            // Retorna el tipus de l'objecte que conté el tipus genèric
            // ModelCollectionBase<> a la seva herència.
            Type type = ReflectionHelper.FindParentType(
                collection.GetType(),
                t => t.IsGenericType
                    && typeof(ModelCollectionBase<>)
                    .Equals(t.GetGenericTypeDefinition())
                    );
            // Si s'ha trobat el tipus a l'herència procedeix.
            if (type != null)
            {
                // Obté el mètode de la col·lecció.
                MethodInfo methodInfo = type.GetMethod(methodName);
                // Si s'ha trobat el mètode s'invoca.
                if (methodInfo != null)
                    // S'invoca el mètode.
                    methodInfo.Invoke(collection, null);
            }
        }

        #endregion

        #region Helper Method CanExecuteModelCollectionCommand

        /// <summary>
        /// Condició d'execució de commanda que s'extreu d'una propietat d'un objecte
        /// <see cref="ModelCollectionBase" /> mitjançant <see cref="Reflection"/>.
        /// </summary>
        /// <param name="collection">Paràmetre de la commanda.</param>
        /// <param name="methodName">Nom de la propietat amb la condició.</param>
        /// <returns>Autorització d'execució de la commanda.</returns>
        private bool CanExecuteModelCollectionCommand(object collection, string propertyName)
        {
            // Si el paràmetre de la comanda és nul surt.
            if (collection == null)
                return false;
            // Retorna el tipus de l'objecte que conté el tipus genèric
            // ModelCollectionBase<> a la seva herència.
            Type type = ReflectionHelper.FindParentType(
                collection.GetType(),
                t => t.IsGenericType
                    && typeof(ModelCollectionBase<>)
                    .Equals(t.GetGenericTypeDefinition())
                    );
            // Si s'ha trobat el tipus a l'herència procedeix.
            if (type != null)
            {
                // Obté la propietat de la col·lecció.
                PropertyInfo propInfo = type.GetProperty(propertyName);
                // Si s'ha trobat la propietat s'obté si hi ha canvis.
                if (propInfo != null)
                {
                    // Obté la propietat de la col·lecció.
                    object value = propInfo.GetValue(collection);
                    // Si s'ha trobat la propietat en retorna el valor boleà.
                    return value as bool? ?? false;
                }
            }
            // Si no s'ha trobat el tipus d'herència retorna fals.
            return false;
        }

        #endregion

        #endregion

    }

}
